/*
 * Copyright European Organization for Nuclear Research (CERN)
 * All rights reserved
 *
 * Author:	Vasilis.Vlachoudis@cern.ch
 * Date:	27-Feb-2023
 */
#ifndef __DATA_MESHDATA_H
#define __DATA_MESHDATA_H

#include <assert.h>

#include <fstream>
#include <iostream>

#include "math/matrix4.h"

// ================================= MeshData ==================================
/* 3D mesh data on various grid formats
 *  - Rectilinear X-Y-Z
 *  - Cylindrical R-Phi-Z
 * With possible symmetries
 *
 * If the standard system is used
 *    X = horizontal
 *    Y = vertical
 *    Z = horizontal-beam direction
 * Then the transformation that has to be performed are
 *     X -> -X      [X]  Y   Z     Size: -1.0
 *     Y ->  Z       X   Y  [Z]    Size:  1.0
 *     Z ->  Y       X  [Y]  Z     Size:  1.0
 */
class MeshData {
protected:
	std::string	_filename;
	std::string	_name;		// name of detector loaded
	int		_detector;	// detector loaded

public:					// FIXME move to protected
	int	nx,    ny,    nz;
	double	xlow,  ylow,  zlow;
	double	xhigh, yhigh, zhigh;
	double	dx,    dy,    dz;
	//Point	offset;
	double	xofs,  yofs,  zofs;	// offset of usrbin
	double	x0,  y0;		// for R-Z-Phi binning
	double	min, max;

protected:
	// flags of type of scoring
	bool	xsym;			// symmetry on X / R
	bool	ysym;			// symmetry on Y / Phi
	bool	zsym;			// symmetry on Z
	bool	rphizMesh;		// cylindrical
	bool	regionMesh;		// per region data

	double	_norm;
	bool	_logscale;		// is logarithmic histogram
	bool	_checker;
	bool	_hasMatrix;		// if rotation matrix is set
	Matrix4	_matrix;

	float*	 data;			// FIXME convert to double
	float*	_error;			// FIXME convert to double
	bool	_errorsPresent;
//	double	 _total;
//
private:
	time_t	_mtime;			// last modified time

public:
	MeshData()
		: data(nullptr),
		  _error(nullptr),
		  _errorsPresent(false)
		{ cleanup(); }
	//MeshData(const char* fn, int det) : data(NULL) { load(fn,det); }
	MeshData(const MeshData& src)
		: data(nullptr),
		  _error(nullptr),
		  _errorsPresent(false)
		{ copy(src); }
virtual	~MeshData()	{ cleanup(); }

virtual	void	cleanup();
virtual	bool	load(const char* /*fn*/, int /*det*/) { return true; }
virtual	bool	save(const char* /*fn*/=NULL)	{ return true; }
	void	copy(const MeshData& src);

	bool	errorsPresent() const { return _errorsPresent; };

	/* create usrbin */
	//void	create(int det, Type atype, int ascore,
	//		const double axlow, const double axhigh, const int anx,
	//		const double aylow, const double ayhigh, const int any,
	//		const double azlow, const double azhigh, const int anz);
	void	checker(const bool rphiz,
			double axlow, double axhigh, const int anx,
			double aylow, double ayhigh, const int any,
			double azlow, double azhigh, const int anz);
	bool	checker()	const	{ return _checker; }

	/* get/set */
	int	detector()	const	{ return _detector; }
	std::string name()	const	{ return _name; }
	void	name(std::string n)	{ _name = n; }
	std::string filename()	const	{ return _filename; }
	bool	isregion()	const	{ return regionMesh; }
	int	totalSize()	const	{ return nx * ny * nz; }
	bool	hasData()	const	{ return data != NULL; }
	double	xmin()		const	{ return xsym? -xhigh : xlow; }
	double	ymin()		const	{ return ysym? -yhigh : ylow; }
	double	zmin()		const	{ return zsym? -zhigh : zlow; }
	double	xmax()		const	{ return xhigh; }
	double	ymax()		const	{ return yhigh; }
	double	zmax()		const	{ return zhigh; }

	void	norm(const double n);
	double	norm()			const	{ return _logscale? exp10(_norm) : _norm; }
	void	normalize(const double n);

	bool	convert(bool tolog);
	bool	logscale()		const	{ return _logscale; }

	// WARNING no transformation is applied!!!
	int	usrbini(const double x)	const	{ return Int((x - xlow)/dx); }
	int	usrbinj(const double y)	const	{ return Int((y - ylow)/dy); }
	int	usrbink(const double z)	const	{ return Int((z - zlow)/dz); }

	double	usrbinx(const int i)	const	{ return xlow + i*dx; }
	double	usrbiny(const int j)	const	{ return ylow + j*dy; }
	double	usrbinz(const int k)	const	{ return zlow + k*dz; }

	double	usrbincx(const int i)	const	{ return xlow + i*dx + dx/2.0; }
	double	usrbincy(const int j)	const	{ return ylow + j*dy + dy/2.0; }
	double	usrbincz(const int k)	const	{ return zlow + k*dz + dz/2.0; }

	// with transformation
	bool	xyz2ijk(double x, double y, double z, int* i, int* j, int* k) const;

	//   getData returns the normalized data as stored in the memory
	//   WARNING if _logscale is set, it returns the logarithm of the data
	double	getData(const int ptr, bool getError=false) const {
				if (!getError)
					return _normValue(data[ptr]);
				else
					return getErr(ptr);
			}

	/** @return error at ptr */
	double	getErr(const int ptr) const {
				assert(_error);
				if (_errorsPresent)
					return _error[ptr];
				return 0.0;
			}

	double	getData(int i, int j, int k, bool getError=false) const {
				if (!getError)
					return _normValue(data[(k*ny + j)*nx + i]);
				else
					return getErr(i, j, k);
			}

	/** @return error at location i,j,k */
	double	getErr(int i, int j, int k) const {
				assert(_error);
				if (!_errorsPresent)
					return 0.0;
				return (_error[(k*ny + j)*nx + i]);
			}

	bool	getData(double x, double y, double z, double* value, bool getError=false) const;
	bool	getData(const int region, double* value, bool getError=false) const;
	bool	getDataRPhiZ(double r, double phi, double z, double* value, bool getError=false) const;

	/** set error at location i,j,k */
	void	setErr(int i, int j, int k, double value) {
				assert(_error);
				_error[(k*ny + j)*nx + i] = (float)value;
				_errorsPresent = true;
			}

	/** @return real normalized value for cell indexed ptr */
	double	get(const int ptr, bool getError=false) const {
				if (_logscale)
					return exp10(getData(ptr, getError));
				else
					return getData(ptr, getError);
			}

	/** @return real normalized value at location i,j,k */
	double	get(int i, int j, int k, bool getError=false) const	{
				if (_logscale)
					return exp10(getData(i,j,k, getError));
				else
					return getData(i,j,k, getError);
			}

	/** @return real normalized value at location */
	bool	get(double x, double y, double z, double* value, bool getError=false) const {
				bool ok = getData(x,y,z,value, getError);
				if (ok && _logscale)
					*value = exp10(*value);
				return ok;
			}

	/** @return real normalized value at location given by point P */
	double	get(const Point& P, bool getError=false) const {
				double value;
				if (get(P.x, P.y, P.z, &value, getError)) {
					return value;
				} else
					return 0.0;
			}

	/** @return real normalized value for the region */
	bool	get(const int region, double* value, bool getError=false) const {
				bool ok = getData(region, value, getError);
				if (ok && _logscale)
					*value = exp10(*value);
				return ok;
			}

	/** set a value to a cell */
	void	set(int i, int j, int k, double value, bool setError=false) {
				assert(!_logscale && !_checker && data);
				// NOTE: due to the current float array we must
				// cast when assigning. to be fixed and changed
				// to double
				if (!setError)
					data[(k*ny + j)*nx + i] = (float)value;
				else {
					assert(_error);
					_error[(k*ny + j)*nx + i] = (float)value;
				}
			}

	/** increase cell value by amount v */
	void	add(int i, int j, int k, float value, bool addError=false) {
				assert(!_logscale && !_checker && data);
				if (!addError)
					data[(k*ny + j)*nx + i] += value;
				else {
					assert(_error);
					_error[(k*ny + j)*nx + i] += value;
					_errorsPresent = true;
				}
				/*_total += value;*/
			}
	void	add(double x, double y, double z, float value, bool addError=false) {
				assert(!_logscale && !_checker && data);
				int i,j,k;
				if (xyz2ijk(x,y,z, &i, &j, &k))
					add(i,j,k,value, addError);
			}
	double	add(Point A, Point B, double value=1.0, bool tracklength=true);

	/** @return integral on a straight line */
	double	integrate(Point A, Point B, double t) const;

	void	offset(const Point& ofs);
	void	scanMinMax();

	// transformation matrix operations
	void	clearMatrix()			{ _hasMatrix = false; }
	bool	hasMatrix()		const	{ return _hasMatrix; }
	void	matrix(const Matrix4& M);
const	Matrix4& matrix()		const	{ return _matrix; }

protected:
	int	preLoad(const char* fn, int det);
	void	postLoad();
virtual void	allocateData()  { data  = new float[totalSize()]; }
virtual void	allocateError() {
				_error = new float[totalSize()];
				memset(_error, 0, sizeof(*_error) * totalSize());
			}

private:
	// returned normalized value
	double	_normValue(double value) const {
				if (_logscale)
					return value + _norm;
				else
					return value * _norm;
			}

	double	intersectXYZ(Point& P, const Vector& D) const;
	void	_xyzStepInit(double pos, double dir, double low, double high,
				double* t, double* dt, int* step, int* out) const;

	// std dump operator via stream inserter
friend	std::ofstream& operator<<(std::ofstream&, const MeshData&);
}; // MeshData

#endif
