////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// F A R M     -  Flair advanced rendering module
//
// Copyright European Organization for Nuclear Research( CERN )
// All rights reserved
//
// Authors: Christian.Theis@cern.ch
//			    Vasilis.Vlachoudis@cern.ch
//
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

#pragma once
#ifndef ARCHI_INTEGRATOR_H
#define ARCHI_INTEGRATOR_H

#include "baseintegrator.h"
#include "farmsurfaceintegrator.h"
#include "graph/color.h"
#include "matphong.h"
#include "lightmap.h"
#include "biasedlightmap.h"
#include "geometry/material.h"
#include "mattoon.h"

#include <vector>

class D3Layer;

class CArchiIntegrator : public CBaseIntegrator
{
public:
	CArchiIntegrator();
	virtual ~CArchiIntegrator( void );

	// initialize options configured in the optional config object in FARM
	virtual void InitFromConfig();

	// main integrator routine returning the luminance for the passed ray & rendering pass
	virtual Color Li( GeometryEngine* pEngine, Ray* pRay, const int RenderPass, int Light0, int CallerId = CBaseIntegrator::cStandard );

	// configure ambient occlusion
	void SetAOCParameters( double MinIntensity, double MaxIntensity, double MaxDist, int NrOfSamples, int Multiplier )
	{
		m_AOCMinIntensity = MinIntensity;
		m_AOCMaxIntensity = MaxIntensity;
		m_AOCMaxDist      = MaxDist;
		m_NrOfAOCSamples  = NrOfSamples;
		m_AOCSampleMultiplier = Multiplier;
	}

	void SetAOCSamples(int NrOfSamples) { m_NrOfAOCSamples = NrOfSamples; }
	int  GetAOCSamples() const          { return m_NrOfAOCSamples; }

	// retrieve the light intensity at the hit point -> used to apply shadows for data that is displayed
	double	GetLightIntensityAtHit( GeometryEngine* pEngine, Ray* pRay, const int RenderPass, int Light0 ) { auto RetVal = ShadeHitPoint( pEngine, pRay, RenderPass, Light0 ); return RetVal.luminance(); };

protected:

	// ambient occlusion integrator
	double CalcAmbientOcclusion( GeometryEngine* eng, Ray* pixelRay, int nrOfSamples, const double cMaxDist, const double cMinIntensity, const double cMaxIntensity );

	// refraction & transmission
	virtual Color Li_Transmission( GeometryEngine* pEngine, Ray* pRay, const int RenderPass, int Light0 );

	// calculate shading intensity of the current hit point
	virtual Color ShadeHitPoint( GeometryEngine* pEngine, Ray* pRay, const int RenderPass, int Light0 );

	// set up and get refraction ray
	void				  GetRefractionRay( const Ray& Out, Ray& In, const Vector& Normal );

	// set up shaders (if needed)
	virtual void SetupBxDFLUT() {};

	// get base hit color
	virtual Color GetHitColor( Ray* pRay ) { UNUSED( pRay );  return Color::White; };

protected:
	double		m_AOCMinIntensity;
	double		m_AOCMaxIntensity;
	double		m_AOCMaxDist;
	int				m_NrOfAOCSamples;
	int				m_MaxAOCDepth;
	int				m_AOCSampleMultiplier;

	bool			m_bTexAvailable;

	std::vector<CPhysMatInterface*> m_PhysMatLUT;
};

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// Colored architectural renderer
//
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
class CColoredArchiIntegrator : public CArchiIntegrator
{
public:
	CColoredArchiIntegrator()
	{
		m_AOCMaxIntensity = 0.5;
	}

	virtual ~CColoredArchiIntegrator( void ) {
		for ( size_t i = 0; i < m_PhysMatLUT.size(); ++i )
			delete m_PhysMatLUT[i];

		m_PhysMatLUT.clear();
	}

	// initialize options configured in the optional config object in FARM
	virtual void InitFromConfig() override
	{
		CArchiIntegrator::InitFromConfig();
	}


protected:
	void SetupBxDFLUT() override {
		m_PhysMatLUT.push_back( new CMatToon( this ) );
	}

	// shade hit point
	Color ShadeHitPoint( GeometryEngine* pEngine, Ray* pRay, const int RenderPass, int Light0 ) override;

	// get hit color
	Color GetHitColor( Ray* pRay ) override { return m_PhysMatLUT[0]->GetHitColor( pRay ); }; // get base hit color
};

#endif
// vim:ts=2
