////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// F A R M     -  Flair advanced rendering module
//
// Copyright European Organization for Nuclear Research( CERN )
// All rights reserved
//
// Authors: Christian.Theis@cern.ch
//			    Vasilis.Vlachoudis@cern.ch
//
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

#include "biasedlightmap.h"
#include "bxdfbase.h"
#include <vector>
#include <iostream>
#include <fstream>

using namespace std;


CBiasedLightMap::CBiasedLightMap()
{
}

CBiasedLightMap::~CBiasedLightMap(void)
{
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

bool CBiasedLightMap::LoadMap( const std::string& FileName, Geometry* pGeometry )

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
  bool bSuccess = CLightMap::LoadMap( FileName, pGeometry );
  if( !bSuccess )
    return false;
  else
    CalculateBiasingDistributions();

  return bSuccess;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

void  CBiasedLightMap::CalculateBiasingDistributions()

//
// The basic idea is that the image map is seen as a 2D map with X = U and Y = V.
// First we calculate the integral CDF for each column V in steps of m_IntegralStep
// pixels. Thus, we have an array which keeps the tabulated CDF for each column.
// Furthermore, we calculate on integral CDF in horizontal direction which integrates
// the values of each integrated column CDF.
//
// The sampling procedure first samples the horizontal U CDF uniformly to determine
// in which column we will sample. As soon as the row is determined we sample the respective
// column uniformly to determine the V bin, in which we sample uniformly to get the actual
// texel position.
//
// [5/26/2020 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
  int             Width   = m_pLightMap->Width();
  int             Height  = m_pLightMap->Height();
  double          CurColCDFValue = 0.0;
  double          CurRowCDFValue = 0.0;
  double          CurLuminance;
  vector<double>  CurColumnCDF;
  Color           PixelColor;
  COLORREF        PixelRGB;

  // clear everything
  vector<vector<double> >().swap( m_ColumnCDF);
  vector<double>().swap( m_IntegralRowCDF );

  // We walk over all columns and calculate the integral CDF for each column and save it in a table.
  // Furthermore, we calculate the integral tabulated row CDF which sums up the integral value of each column
  for( int x = 0; x < Width; ++x ) {
    for( int y = 0; y < Height; ++y ) {

      PixelRGB = m_pLightMap->GetPixel( x, y);
      PixelColor = Color( GetRValue( PixelRGB) / 255.0, GetGValue( PixelRGB)  / 255.0, GetBValue( PixelRGB) / 255.0 );
      CurLuminance = PixelColor.luminance();

      CurColCDFValue += CurLuminance;

      // save the values in a table with a lower resolution than the actual image height to save some time
      if( !(y % m_IntegralStep) )
        CurColumnCDF.push_back( CurColCDFValue );

    }
    CurColumnCDF.push_back( CurColCDFValue );     // don't forget the last bin which might not have fit in Height % IntegralStep

    // normalize column CDF to 1 & save it
    double Integral = *(CurColumnCDF.rbegin());
    for( size_t i = 0; i < CurColumnCDF.size(); ++i ) {
      CurColumnCDF[i] = CurColumnCDF[i] / Integral;
    }

    m_ColumnCDF.push_back( CurColumnCDF );                // save the tabulated CDF for each column!

    CurRowCDFValue += CurColCDFValue;

    // save the values in a table with a lower resolution than the actual image width to save some time
    if( !(x % m_IntegralStep) )
      m_IntegralRowCDF.push_back( CurRowCDFValue );

    // reset for next column
    CurColCDFValue = 0.0;
    vector<double>().swap( CurColumnCDF );
  }
  m_IntegralRowCDF.push_back( CurRowCDFValue ); // don't forget the last bin which might not have fit in Width % IntegralStep


  // normalize row CDF to 1 & save it
  double Integral = *(m_IntegralRowCDF.rbegin());
  for( size_t i = 0; i < m_IntegralRowCDF.size(); ++i ) {
    m_IntegralRowCDF[i] = m_IntegralRowCDF[i] / Integral;
  }

  // DEBUG
//   ofstream Out( "lightmap-cdf.dat" );
//   Out << "Row CDF" << endl;
//   for ( auto Val : m_IntegralRowCDF ) {
//     Out << Val << ", ";
//   }
//   Out << endl << endl << endl;
// 	Out << "COL CDF" << endl;
// 	for ( size_t i = 0; i < m_ColumnCDF.size(); ++i ) {
//     Out << "Col " << i << endl;
//     Out << "-----------------" << endl;
//     for ( size_t j = 0; j < m_ColumnCDF[i].size(); ++j ) {
//       Out << m_ColumnCDF[i][j] << ", ";
//     }
//     Out << endl;
// 	}

}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

void CBiasedLightMap::SampleDirectionUV( const Vector& Normal, Random& Sampler, unsigned int SampleNr, Vector& Dir, float& UTexel, float& VTexel )

// The sampling procedure first samples the horizontal U CDF uniformly to determine
// in which column we will sample. As soon as the row is determined we sample the respective
// column uniformly to determine the V bin, in which we sample uniformly to get the actual
// texel position.
//
// [5/26/2020 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
  UNUSED( SampleNr ); // will be needed for QMC

  // TODO: replace with QMC
  double  USample = Sampler.real();
  double  VSample = Sampler.real();
  size_t  ColIndex = 0, RowIndex = 0;
  size_t  MinSearchIndex = 0;
  size_t  MaxSearchIndex = 0;

  // get column index
  /* LEGACY code: using linear search
  for( ColIndex  = 0; ColIndex < m_IntegralRowCDF.size() - 2; ++ColIndex ) {
    if( USample >= m_IntegralRowCDF[ColIndex] && USample < m_IntegralRowCDF[ColIndex + 1] )
      break;
  }
  */

  // New code using binary search for the lookup
  MinSearchIndex = 0;
  MaxSearchIndex = m_IntegralRowCDF.size() - 1;
  ColIndex       = BinSearchIndex( &m_IntegralRowCDF[0], USample, MinSearchIndex, MaxSearchIndex );

  // convert the column table index to an actual pixel position by sampling uniformly within the bin which
  // contains several columns
  USample = ColIndex * m_IntegralStep + Sampler.real() * m_IntegralStep;
  if( USample > (m_pLightMap->Width() - 1.0) )
    USample = m_pLightMap->Width() - 1.0;
  USample = std::max<double>( 0.0, std::min<double>( USample, m_ColumnCDF.size() - 1 ) );

  // get row index in the column indicated by ColIndex
  /* LEGACY code: using linear search
  for( RowIndex = 0; RowIndex < m_ColumnCDF[(int)USample].size() - 2; ++RowIndex ) {
    if( VSample >= m_ColumnCDF[(int)USample][RowIndex] && VSample < m_ColumnCDF[(int)USample][RowIndex + 1] )
      break;
  }*/

  // New code using binary search for the lookup
  MinSearchIndex = 0;
  MaxSearchIndex = m_ColumnCDF[(int)USample].size() - 2;
  RowIndex = BinSearchIndex<double>( &m_ColumnCDF[(int)USample][0], VSample, MinSearchIndex, MaxSearchIndex );

  VSample = RowIndex * m_IntegralStep + Sampler.real() * m_IntegralStep;
  if( VSample > (m_pLightMap->Height() - 1.0) )
    VSample = m_pLightMap->Height() - 1.0;
  VSample = std::max( 0.0, std::min<double>( VSample, m_IntegralRowCDF.size() ) );

  UTexel = USample;
  VTexel = VSample;

  // convert the U & V positions to cartesian coordinates via spherical coordinates, assuming U => phi [0, 2pi] and V => theta [-pi/2, pi/2]
  double Phi   = (UTexel / m_pLightMap->Height()) * 2 * PI;
  double Theta = -PIover2 + (VTexel / m_pLightMap->Width()) * PI;

  Dir.x = sin( Theta ) * cos( Phi );
  Dir.y = sin( Theta ) * cos( Phi );
  Dir.z = cos( Theta );
  Dir.normalize();

  // Make sure that our face normal & the sampled direction are in the same half-space. Otherwise we might end up
  // with incorrect self-shadowing. (CT 14.7.2008)
  if( Dir.dot( Normal) < 0.0 ) {
    Dir = Dir * -1.0;
  }

  // transform from canonical system (Z = up = V direction) to
  // the requested coordinate system
  Dir = m_Projection * Dir;

}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

Color CBiasedLightMap::LookUpColor( float UTexel, float VTexel )

// [5/26/2020 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
/*
  float u, v;

  u = UTexel - (unsigned int)UTexel;
  v = VTexel - (unsigned int)VTexel;

  u *= u * (m_LightMap.Width()  - 3);
  v *= v * (m_LightMap.Height() - 3);

  int iu = max( 0, min( static_cast<int>(u), m_LightMap.Width() - 2) );
  int iv = max( 0, min( static_cast<int>(v), m_LightMap.Height() - 2) );

  float tu = u - iu;
  float tv = v - iv;
*/
  COLORREF ResColor = m_pLightMap->GetPixel( (unsigned int)UTexel, (unsigned int)VTexel );     // no bilinear interpolation
  return Color( GetRValue( ResColor) / 255.0, GetGValue( ResColor)  / 255.0, GetBValue( ResColor) / 255.0 );

}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

Color CBiasedLightMap::Sample_f( GeometryEngine* pEngine, const Ray& Out, const Vector& Normal, int SampleNr, Vector& LightDir )

// Overloaded biased sampling of color & light direction
//
// [5/27/2020 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
  float UTexel = 0.f, VTexel = 0.f;
  UNUSED( Out );

	// sample the direction and the associated U & V coordinates in the lightmap.
	SampleDirectionUV( Normal, pEngine->random, SampleNr, LightDir, UTexel, VTexel );
	return LookUpColor( UTexel, VTexel );
}