////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// F A R M     -  Flair advanced rendering module
//
// Copyright European Organization for Nuclear Research( CERN )
// All rights reserved
//
// Authors: Christian.Theis@cern.ch
//			    Vasilis.Vlachoudis@cern.ch
//
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

#include "bxdfspectransmission.h"
#include "bxdffresnel.h"

#include <algorithm>

CBxDFSpecTransmission::CBxDFSpecTransmission( double EtaT /* = 0.0*/, const Color& TransmissionColor /* = Color( 1.0, 1.0, 1.0 ) ) */) : CBxDFBase(static_cast<BxDFType>(BSDF_TRANSMISSION | BSDF_SPECULAR)),
                                                 m_pFresnel( new CFresnelDiel), m_TransmissionColor( TransmissionColor ), m_EtaT( EtaT )
{

}

CBxDFSpecTransmission::~CBxDFSpecTransmission(void)
{
  delete m_pFresnel;
}


////////////////////////////////////////////////////////////////////////////////////////////////////////////

Color CBxDFSpecTransmission::Sample_f( const Ray& Out, Ray& In, float& Pdf, const Vector& Normal, const Material& RenderMat, double* pRandNr, BxDFType Type, const Color& LightColor /*= Color( 1.0, 1.0, 1.0 )*/, double Power /*= 1.0*/, float Shade /*= 1.0 */ )

////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
	UNUSED( Power ); UNUSED( LightColor ); UNUSED( Shade );

  if( Type == CBxDFBase::BSDF_TRANSMISSION )
    return Sample_f_Transmission( Out, In, Pdf, Normal, RenderMat, pRandNr, LightColor );

  return Color::Black;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////

Color CBxDFSpecTransmission::Sample_f_Transmission( const Ray& Out, Ray& In, float& Pdf, const Vector& Normal, const Material& RenderMat, double* pRandNr, const Color& LightColor )

////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
	UNUSED( pRandNr );

	bool   bEntering = Out.type == Ray::RayType::Transmission ? false : true;
	double EtaI = 1.0;			// to be physically precise this IOR should be carried by the outgoing ray --> TODO:
	double EtaT = m_EtaT == 0.0 ? RenderMat.ior() : m_EtaT;

	Pdf = 1.0;

	double Cosi = Clamp( Normal.dot( Out.dir() ), -1.0, 1.0 );
	if ( bEntering ) {
		Cosi = -Cosi;
	}
	else {
		std::swap( EtaI, EtaT );
	}

	double k = 1 - (EtaI / EtaT) * (EtaI / EtaT) * (1.0 - Cosi * Cosi);
	if ( k < 0 ) {
		return Sample_f_Reflection( Out, In, Pdf, Normal, RenderMat, pRandNr, LightColor );
	}

	Vector Temp = (EtaI / EtaT) * Out.dir() + ((EtaI / EtaT) * Cosi - sqrtf( k )) * Normal;
	Temp.normalize();


	// set up ray
	In = Out;
	In.init();
	In.depth = Out.depth + 1;

	RaySegment NewSegment( Out.hit( SMALL3D4 * 100 ), Temp, Out.hitZone() );
	NewSegment.tmin = SMALL3D3;
	NewSegment.tmax = INFINITE;

	In.push( NewSegment );
	In.type = bEntering ? Ray::RayType::Transmission : Ray::RayType::Standard;	// if the out ray was entering then this is a transmission ray (inside a solid), otherwise it is an exiting ray

	// Fresnel transmission,
	double Rho = 1.0 / fabs( Cosi ) * (EtaI * EtaI) / (EtaT * EtaT) * (1.0 - m_pFresnel->Evaluate( Cosi, EtaI, EtaT ));
	return m_TransmissionColor * LightColor * Rho;

}

////////////////////////////////////////////////////////////////////////////////////////////////////////////

Color CBxDFSpecTransmission::Sample_f_Reflection( const Ray& Out, Ray& In, float& Pdf, const Vector& Normal, const Material& RenderMat, double* pRandNr, const Color& LightColor )

////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
	UNUSED( pRandNr );

	/////
	double EtaI = RenderMat.ior();	// NOTE: we have to reverse EtaI & EtaT as this function is only triggered for total reflection!
	double EtaT = 1.0;
	double ReflectionCoeff = 0;

	Pdf = 1.0f;

	// CT: we only render reflections if we are at the beauty pass (step = 1)
	double dot = Normal * Out.dir();
	if ( dot < 0.0 ) {
		Vector reflect = (-2.0 * dot) * Normal + Out.dir();
		reflect.normalize();

		// Find intersection with light
		In = Out;
		In.init();
		In.depth = Out.depth + 1;
		In.push( RaySegment( Out.hit( -SMALL3D4 ), reflect, Out.prevZone() ) );

		In.type = Ray::RayType::Transmission;

		ReflectionCoeff = m_pFresnel->Evaluate( fabs( dot ), EtaI, EtaT );
		return LightColor( ReflectionCoeff * RenderMat.specular() / fabs( dot ) );
	}

	return Color::Black;
}

// vim:ts=2
