////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// F A R M     -  Flair advanced rendering module
//
// Copyright European Organization for Nuclear Research( CERN )
// All rights reserved
//
// Authors: Christian.Theis@cern.ch
//			    Vasilis.Vlachoudis@cern.ch
//
// Based on SimpleGeo's CDIB class from 2003, ported from my 3D particle simulation system (PSG) 1998
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

#include "dib.h"

#include <cmath>
#include <cstring>
#include <iostream>
#include <fstream>
#include <sstream>
#include <zlib.h>

#define BIG_DISTANCE 10000000L

#define DIST(r1,g1,b1,r2,g2,b2) \
	    (long) (3L*(long)((r1)-(r2))*(long)((r1)-(r2)) + \
		    4L*(long)((g1)-(g2))*(long)((g1)-(g2)) + \
		    2L*(long)((b1)-(b2))*(long)((b1)-(b2)))


static unsigned char masktable[] = { 0x80,0x40,0x20,0x10,0x08,0x04,0x02,0x01 };

using namespace std;

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

CDIB::CDIB()

// [3/11/2020 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
	m_pVoid			 = NULL;
	m_pLinePtr	 = NULL;
	m_bUseGamma	 = false;
	width=height = 0;
	m_Name = "unnamed";
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

CDIB::~CDIB()

// [3/11/2020 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
	DestroyDIB();
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

void CDIB::DestroyDIB()

// [3/11/2020 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
	if(m_pVoid)
		free(m_pVoid);

	m_pVoid = NULL;
	if(m_pLinePtr)
		free(m_pLinePtr);

	m_pLinePtr = NULL;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

bool CDIB::Create(int Width, int Height, int Bits)

// [3/11/2020 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
	/*
		Free existing image
	*/
	DestroyDIB();
//	assert(bits == 24 || bits == 8);

	BITMAPINFOHEADER bmInfo;

	memset(&bmInfo,0,sizeof(BITMAPINFOHEADER));
	bmInfo.biSize					= sizeof(BITMAPINFOHEADER);
	bmInfo.biWidth				= Width;
	bmInfo.biHeight				= Height;
	bmInfo.biPlanes				= 1;
	bmInfo.biBitCount			= Bits;
	bmInfo.biCompression	= BI_RGB;
	return Create(bmInfo);
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

bool CDIB::Create(BITMAPINFOHEADER& bmInfo)

// [3/11/2020 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
	bytes = (bmInfo.biBitCount*bmInfo.biWidth)>>3;
	height = bmInfo.biHeight;
	width = bmInfo.biWidth;
//	bmInfo.biHeight *= -1;
	while(bytes%4) bytes++;

	int size;
	size = sizeof(BITMAPINFOHEADER) + sizeof(RGBQUAD)*GetPaletteSize(bmInfo) + bytes*height;
	m_pVoid = (void *)malloc(size);
	if(!m_pVoid) return false;

	m_pInfo = (BITMAPINFO*)m_pVoid;
	memcpy((void *)&m_pInfo->bmiHeader,(void *)&bmInfo,sizeof(BITMAPINFOHEADER));
	m_pRGB = (RGBQUAD *)((unsigned char *)m_pVoid + sizeof(BITMAPINFOHEADER)) ;
	m_pBits = (unsigned char *)(m_pVoid) + sizeof(BITMAPINFOHEADER) + sizeof(RGBQUAD)*GetPaletteSize();

	int i;
	unsigned char **ptr;
	m_pLinePtr = (unsigned char **)malloc(sizeof(unsigned char *)*height);
	if(!m_pLinePtr) return false;
	for(i=0,ptr=m_pLinePtr; i < height; i++,ptr++) {
		//*ptr = (int)(m_pBits)+(i*bytes);
		//*ptr = (int)GetLinePtr(i);
		*ptr = m_pBits + (height-i-1)*bytes;
	}
	m_nFlags = 0;
	return true;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

void CDIB::SetPalette(unsigned char *palette)

// [3/11/2020 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
	int i,size;
	RGBQUAD *rgb;
	if(!palette) return;
	size = GetPaletteSize();
	for(i=0,rgb = m_pRGB; i < size; i++,rgb++,palette+=3)
	{
		if(m_bUseGamma)
		{
			rgb->rgbRed = Gamma[palette[0]];
			rgb->rgbGreen = Gamma[palette[1]];
			rgb->rgbBlue = Gamma[palette[2]];
		}
		else
		{
			rgb->rgbRed = palette[0];
			rgb->rgbGreen = palette[1];
			rgb->rgbBlue = palette[2];
		}
		rgb->rgbReserved = (unsigned char)0;
	}
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

void CDIB::SetPalette(RGBQUAD *pRGB)

// [3/11/2020 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
int size;
	if(!pRGB) return;
	size = GetPaletteSize();
	memcpy(m_pRGB,pRGB,size*sizeof(RGBQUAD));
}


////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

int CDIB::GetPaletteSize()

// [3/11/2020 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
	return GetPaletteSize(m_pInfo->bmiHeader);
}


////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

int CDIB::GetPaletteSize(BITMAPINFOHEADER& bmInfo)

// [3/11/2020 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
	switch(bmInfo.biBitCount)
	{
	case 1:
			return 2;
	case 4:
			return 16;
	case 8:
			return 256;
	default:
			return 0;
	}
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

void CDIB::SetPixel(int x,int y, unsigned char Red, unsigned char Green, unsigned char Blue)

// [3/11/2020 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
unsigned char *ptr;
	assert(x >= 0 && y >=0);
	assert(x < width && y < height);

	ptr = (unsigned char *)m_pLinePtr[y];
	ptr += x * 4;

  *ptr++ = Blue;
	*ptr++ = Green;
	*ptr++ = Red;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

void CDIB::SetPixel(int x, int y, COLORREF color)

// [3/11/2020 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
unsigned char *ptr;
	assert(x >= 0 && y >=0);
	assert(x < width && y < height);

//	ptr = m_pBits + (y*bytes) + x * 3;
	ptr = (unsigned char *)m_pLinePtr[y];
	ptr += x*3;
	*ptr++ = (unsigned char)GetBValue(color);
	*ptr++ = (unsigned char)GetGValue(color);
	*ptr++ = (unsigned char)GetRValue(color);
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

void CDIB::SetPixel8(int x,int y,unsigned char color)

// [3/11/2020 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
	unsigned char *ptr;
	assert(x >= 0 && y >=0);
	assert(x < width && y < height);

//	ptr = m_pBits + (y*bytes) + x ;
//	ptr = (unsigned char *)m_pLinePtr[y] ;
	ptr = GetLinePtr(y);
	ptr += x;
	*ptr = color;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

COLORREF CDIB::GetPixel(int x,int y)

// [3/11/2020 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
	unsigned char *ptr;
	COLORREF color;
	assert(x >= 0 && y >=0);
	assert(x < width && y < height);

	//	ptr = m_pBits + (y*bytes) + x * 3;
	ptr = GetLinePtr(y);
	ptr += (x*3);
	color = MAKE_RGB(*(ptr+2),*(ptr+1),*ptr);
	return color;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

void CDIB::CopyLine(int source,int dest)

// [3/11/2020 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
unsigned char *src,*dst;
	assert(source <= height && source >= 0);
	assert(dest <= height && dest >= 0);
	if(source == dest) return;
	src = GetLinePtr(source);
	dst = GetLinePtr(dest);
	memcpy(dst,src,bytes);
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

void CDIB::InitDIB(COLORREF color)

// [3/11/2020 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
int i,j;
unsigned char *ptr;

	if(m_pInfo->bmiHeader.biBitCount == 24)
	{
		unsigned char col[3];
		col[0]=GetBValue(color);
		col[1]=GetGValue(color);
		col[2]=GetRValue(color);
		for(i=0,ptr = m_pBits; i < height; i++)
		{
			ptr = m_pBits + i*bytes;
			for(j=0; j < width ; j++,ptr+=3)
			{
				memcpy(ptr,col,3);
			}
		}
	}
	else
	{
		for(i=0,ptr = m_pBits; i < height; i++,ptr+=bytes)
		{
			memset(ptr,(unsigned char)color,bytes);
		}
	}
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

void CDIB::ExpandBlt( int nXDest, int nYDest, int xRatio, int yRatio, CDIB& dibSrc, int xSrc, int ySrc, int  nSWidth, int nSHeight )

// [5/25/2020 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
	SetPalette( dibSrc.m_pRGB );

	nSWidth = xSrc + nSWidth > dibSrc.width ? dibSrc.width - xSrc : nSWidth;
	nSHeight = ySrc + nSHeight > dibSrc.height ? dibSrc.height - ySrc : nSHeight;

	Expand( nXDest, nYDest, xRatio, yRatio, dibSrc, xSrc, ySrc, nSWidth, nSHeight );
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

void CDIB::Expand(int nXDest,int nYDest,int xRatio,int yRatio,CDIB& dibSrc,int xSrc,int ySrc,int  nSWidth,int nSHeight)

// [3/11/2020 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
	int xNum,yNum,xErr,yErr;
	int nDWidth,nDHeight;

	nDWidth = nSWidth*xRatio;
	nDHeight = nSHeight*yRatio;

	nDWidth = nXDest+nDWidth > width ? width-nXDest : nDWidth ;
	nDHeight = nYDest+nDHeight > height ? height-nYDest : nDHeight;

	xNum = nDWidth/xRatio;
	yNum = nDHeight/yRatio;
	xErr = nDWidth%xRatio;
	yErr = nDHeight%yRatio;

	unsigned char *buffer,*srcPtr,*destPtr,*ptr;
	int i,j,k = 0;

	buffer = (unsigned char *)malloc(nDWidth+20);
	if(!buffer) return;

	for(i=0; i < yNum; i++,ySrc++) {
		srcPtr = dibSrc.GetLinePtr(ySrc) + xSrc;
		ptr = buffer;
		for(j=0; j < xNum; j++,ptr+=xRatio) {
			memset(ptr,*(srcPtr+j),xRatio);
			k=*(srcPtr+j);
		}
		memset(ptr,(unsigned char)k,xErr);
		for(j=0; j < yRatio ; j++,nYDest++) {
			destPtr = GetLinePtr(nYDest) + nXDest;
			memcpy(destPtr,buffer,nDWidth);
		}
	}
	for(j=0; j < yErr; j++,nYDest++) {
		destPtr = GetLinePtr(nYDest) + nXDest;
		memcpy(destPtr,buffer,nDWidth);
	}
	free(buffer);
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

void CDIB::StretchBlt(int nXDest,int nYDest,int nDWidth,int nDHeight,CDIB& dibSrc,int xSrc,int ySrc,int  nSWidth,int nSHeight)

// [3/11/2020 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
	SetPalette(dibSrc.m_pRGB);
	nDWidth = nXDest+nDWidth > width ? width-nXDest : nDWidth ;
	nDHeight = nYDest+nDHeight > height ? height-nYDest : nDHeight;

	nSWidth = xSrc+nSWidth > dibSrc.width ? dibSrc.width-xSrc : nSWidth;
	nSHeight = ySrc+nSHeight > dibSrc.height? dibSrc.height-ySrc : nSHeight;

	int xDiv,yDiv;
	int xMod,yMod;

	xDiv = nDWidth/nSWidth;
	xMod = nDWidth%nSWidth;

	yDiv = nDHeight/nSHeight;
	yMod = nDHeight%nSHeight;

	if(!xMod && !yMod && xDiv > 0 && yDiv > 0) {
		ExpandBlt(nXDest,nYDest,xDiv,yDiv,dibSrc,xSrc,ySrc,nSWidth,nSHeight);
		return;
	}

	unsigned char *tempPtr,*srcPix,*destPix,*q;
	tempPtr = (unsigned char *)malloc(nDWidth+20);
	int i,j,k,l,x,y,m;
	int xErr,yErr;

	for(i=yErr=m=0; i < nSHeight; i++) {
		srcPix = dibSrc.GetLinePtr(i+ySrc) + xSrc;
		q = tempPtr;
		for(j=l=xErr=0; j < nSWidth; j++,srcPix++) {
			k = xDiv;
			xErr += xMod;
			if(xErr >= nSWidth)
			{
				k++;
				xErr%=nSWidth;
			}
			x=0;
			while(l < nDWidth &&  x < k) {
				*q++ = *srcPix;
				l++;
				x++;
			}
		}
		while(l < nDWidth) {
			*q++=*srcPix;
			l++;
		}
		k= yDiv;
		yErr += yMod;
		if(yErr >= nSHeight) {
			k++;
			yErr%=nSHeight;
		}
		y=0;
		while(m < nDHeight && y < k) {
			destPix = GetLinePtr(m+nYDest) + nXDest;
			memcpy(destPix,tempPtr,nDWidth);
			m++;
			y++;
		}
	}
	while(m < nDHeight ) {
		destPix = GetLinePtr(m+nYDest) + nXDest;
		memcpy(destPix,tempPtr,nDWidth);
		m++;
	}
	free(tempPtr);
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

void CDIB::BitBlt(int nXDest,int nYDest,int nWidth,int nHeight,CDIB& dibSrc,int nSrcX,int nSrcY, unsigned char* colors)

// [3/11/2020 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
	SetPalette(dibSrc.m_pRGB);
	if(nXDest < 0) {
		nSrcX -= nXDest;
		nWidth += nXDest;
		nXDest=0;
	}
	if(nYDest < 0) {
		nSrcY -= nYDest;
		nHeight += nYDest;
		nYDest=0;
	}
	if(nSrcX < 0) {
		nXDest -= nSrcX;
		nWidth += nSrcX;
		nSrcX=0;
	}
	if(nSrcY < 0) {
		nYDest -= nSrcY;
		nHeight += nSrcY;
		nSrcY=0;
	}
	nWidth = nXDest+nWidth > width ? width-nXDest : nWidth ;
	nHeight = nYDest+nHeight > height ? height-nYDest : nHeight;

	nWidth = nSrcX+nWidth > dibSrc.width ? dibSrc.width-nSrcX : nWidth;
	nHeight = nSrcY+nHeight > dibSrc.height? dibSrc.height-nSrcY : nHeight;

	nWidth = max(0,nWidth);
	nHeight = max(0,nHeight);

	int i,k,l,j;
	unsigned char *srcPtr,*destPtr;
	if(!colors) {
		for(i=0,k=nSrcY,l=nYDest; i < nHeight; i++,k++,l++) {
			if(k < 0 || l < 0) {
				continue;
			}
			else  {
				srcPtr = dibSrc.GetLinePtr(k);
				destPtr = GetLinePtr(l);
				memcpy(destPtr+nXDest,srcPtr+nSrcX,nWidth);
			}
		}
	}
	else {
		for(i=0,k=nSrcY,l=nYDest; i < nHeight; i++,k++,l++) {
			if(k < 0 || l < 0) {
				continue;
			}
			else {
				srcPtr = dibSrc.GetLinePtr(k)+nXDest;
				destPtr = GetLinePtr(l)+nSrcX;
				for(j=0; j < nWidth; j++,srcPtr++,destPtr++)
				{
					if(colors[*srcPtr]) *destPtr=*srcPtr;
				}
			}
		}
	}
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

unsigned char *CDIB::GetLinePtr(int line)

// [3/11/2020 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
/*unsigned char *ptr;
	ptr = m_pBits + (height-line-1)*bytes;
	return ptr;*/
	return m_pLinePtr[line];
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

bool CDIB::CopyDIB(const CDIB& dib)

// [3/11/2020 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
	if(Create(dib.m_pInfo->bmiHeader)) {
		SetPalette(dib.m_pRGB);
		memcpy(m_pBits,dib.m_pBits,height*bytes);
		m_Name = dib.GetName();
		return true;
	}
	return false;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

void CDIB::ReplaceColor(unsigned char oldColor,unsigned char newColor)

// [3/11/2020 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
	int i,j;
	unsigned char *ptr;
	for(i=0; i < height; i++) {
		ptr = GetLinePtr(i);
		for(j=0; j < width; j++)	{
			if(ptr[j] == oldColor) ptr[j] = newColor;
		}
	}
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

CDIB& CDIB::operator=(const CDIB& dib)

// [3/11/2020 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
	CopyDIB(dib);
	return *this;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

DWORD CDIB::GetDIBSize()
// [3/11/2020 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
	return sizeof(BITMAPINFOHEADER) + sizeof(RGBQUAD)*GetPaletteSize() + bytes*height;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

void CDIB::UseGamma(float fg,bool bUse)

// [3/11/2020 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
	m_bUseGamma = bUse;
	m_fOldGamma = m_fGamma;
	m_fGamma = fg;
	CreateGammaCurve();
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

void CDIB::CreateGammaCurve()

// [3/11/2020 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
	int i;
	for(i=0;i<256;++i) {
	    Gamma[i]=(int)(255 * pow((float)i/255, m_fGamma) + 0.5f);
	}
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

void CDIB::GetPixel(unsigned int x,unsigned int y,int& pixel)

// [3/11/2020 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
	assert(x < (unsigned int)Width());
	assert(y < (unsigned int)Height());
	if(x >= (unsigned int)Width()) return;
	if(y >= (unsigned int)Height()) return;
	pixel=(GetLinePtr(y))[x];
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

bool CDIB::Make8Bit(CDIB& dib)

// [3/11/2020 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
int nBits;
	assert(Width() == dib.Width());
	assert(Height() == dib.Height());
	nBits = dib.GetBitCount();
	switch(nBits)
	{
	case 1:
		return SwitchFromOne(dib);
		break;
	case 4:
		return SwitchFromFour(dib);
		break;
	case 8:
		return SwitchPalette(dib);
		break;
	case 24:
		return SwitchFrom24(dib);
		break;
	default:
		return false;
	}
	return false;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

bool CDIB::SwitchFromOne(CDIB& dib)

// [3/11/2020 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
	int i,j,w,h;
	unsigned char *sPtr,*dPtr;
	unsigned char cols[2];
	w = Width();
	h = Height();
	memset(CachePtr,0,sizeof(CachePtr));
	cols[0]=ClosestColor(dib.m_pRGB);
	cols[1]=ClosestColor(dib.m_pRGB+1);
	for(i=0; i < h; i++) {
		dPtr = GetLinePtr(i);
		sPtr = dib.GetLinePtr(i);
		for(j=0 ; j < w; j++,dPtr++) {
			if(!(sPtr[j>>3] & masktable[j&7])) *dPtr = cols[0];
			else *dPtr = cols[1];
		}
	}
	return true;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

bool CDIB::SwitchFromFour(CDIB& dib)

// [3/11/2020 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
	int i,n,j,w,h;
	unsigned char *sPtr,*dPtr;
	unsigned char cols[16];
	w = Width();
	h = Height();
	memset(CachePtr,0,sizeof(CachePtr));
	for(i=0; i < 16; i++) {
		cols[i]=ClosestColor(dib.m_pRGB+i);
	}
	for(i=0; i < h; i++) {
		dPtr = GetLinePtr(i);
		sPtr = dib.GetLinePtr(i);
		for(j=0 ; j < w; j++,dPtr++) {
			if(!(j&1)) n = (*sPtr & 0xf0)>>4;
			else {
				n = *sPtr & 0x0f;
				sPtr++;
			}
			*dPtr = cols[n];
		}
	}
	return true;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

bool CDIB::SwitchPalette(CDIB& dib)

// [3/11/2020 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
	int i,j,w,h;
	unsigned char *sPtr,*dPtr;
	unsigned char cols[256];
	w = Width();
	h = Height();
	memset(CachePtr,0,sizeof(CachePtr));
	for(i=0; i < 256; i++) {
		cols[i]=ClosestColor(dib.m_pRGB+i);
	}
	for(i=0; i < h; i++) {
		dPtr = GetLinePtr(i);
		sPtr = dib.GetLinePtr(i);
		for(j=0 ; j < w; j++,sPtr++,dPtr++) {
			*dPtr = cols[*sPtr];
		}
	}
	return true;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

int CDIB::ClosestColor(RGBQUAD *pRgb)

// [3/11/2020 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
	unsigned int dist=BIG_DISTANCE, d, c = 0;
	RGBQUAD *pQuad=m_pRGB;
	int pSize=GetPaletteSize();
	for( int i = 0; i < pSize;i++) {
		if(CachePtr[i])	{
			if(!memcmp((void *)&CacheQuad[i],(void *)pRgb,3)) {
				return i;
			}
		}
	}
	for( int i=0; i < pSize; i++,pQuad++) {
		d = Distance(*pRgb,*pQuad);
		if(!d) {
			CacheQuad[i]=*pRgb;
			CachePtr[i]=1;
			return i;
		}
		if(dist > d) {
			c = i;
			dist = d;
		}
	}
	CacheQuad[c]=*pRgb;
	CachePtr[c]=1;
	return c;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

unsigned int CDIB::Distance(RGBQUAD& rgb1,RGBQUAD& rgb2)

// [3/11/2020 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
	unsigned int d;
	d =  3*(unsigned)((rgb1.rgbRed)-(rgb2.rgbRed))*(unsigned)((rgb1.rgbRed)-(rgb2.rgbRed));
	d += 4*(unsigned)((rgb1.rgbGreen)-(rgb2.rgbGreen))*(unsigned)((rgb1.rgbGreen)-(rgb2.rgbGreen)) ;
	d += 2*(unsigned)((rgb1.rgbBlue)-(rgb2.rgbBlue))*(unsigned)((rgb1.rgbBlue)-(rgb2.rgbBlue));
	return d;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

bool CDIB::OpenDIB( const std::string& csFileName)

// [3/11/2020 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
	m_Name = csFileName;

	FILE* fp = fopen( csFileName.c_str(), "rb" );
	if ( fp ) {
		fclose( fp );
		return OpenBMP( csFileName );
	}
	else {
		string compFileName = csFileName + ".z";
		fp = fopen( compFileName.c_str(), "rb");
		if ( fp ) {
			fclose( fp );
			return OpenCompressedBMP( compFileName );
		}
		else {	// another attempt with .zz suffix
			compFileName = csFileName + ".zz";
			fp = fopen( compFileName.c_str(), "rb" );
			if ( fp ) {
				fclose( fp );
				return OpenCompressedBMP( compFileName );
			}
		}
	}

	return false;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

bool CDIB::SaveDIB( const std::string& csFileName, BitmapType type)

// [3/11/2020 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
	FILE* fp = fopen( csFileName.c_str(), "wb");
	if( !fp ) {
		return false;
	}
	fclose( fp );

	switch(type)
	{
		case BMP:
			return SaveBMP(csFileName);
		default:
			return false;
	}
	return false;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

bool CDIB::SaveBMP( const std::string& csFileName)

// [3/11/2020 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
	BITMAPFILEHEADER bFile;

	FILE* fp = fopen( csFileName.c_str(), "wb" );
	if( !fp ) {
		return false;
	}

	memset( &bFile, 0, sizeof( bFile ) );
	memcpy((void *)&bFile.bfType,"BM",2);

	bFile.bfSize = GetDIBSize() + sizeof(bFile);
	bFile.bfOffBits = sizeof(BITMAPINFOHEADER) + GetPaletteSize()*sizeof(RGBQUAD) + sizeof(BITMAPFILEHEADER);

	fwrite( &bFile, sizeof( bFile ), 1, fp );
	fwrite( m_pVoid, sizeof( m_pVoid ), GetDIBSize(), fp );
	fclose( fp );
	return true;

}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

bool CDIB::ReadBMP( std::istream& input )

//  [1/29/2024 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
	BITMAPFILEHEADER bFile;
	BITMAPINFOHEADER head;

	if(!input ) {
		return false;
	}

	input.read( (char*)&bFile,sizeof(bFile));
	if(memcmp((void *)&bFile.bfType,"BM",2)) {
		//input.close();
		return false;
	}

	input.read( (char*)&head,sizeof(head));
	if(!Create(head)) {
		//input.close();
		return false;
	}

	input.read( (char*)m_pRGB,sizeof(RGBQUAD)*GetPaletteSize());
	input.seekg( bFile.bfOffBits, ios_base::beg );
	input.read( (char*)m_pBits, height * bytes);
	//input.close();
	return true;

}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

bool CDIB::OpenBMP( const std::string& csFileName )

// [5/25/2020 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
	std::ifstream file( csFileName, std::ifstream::binary );

	if ( !file ) {
		return false;
	}

	return ReadBMP( file );
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

bool CDIB::OpenCompressedBMP( const std::string& csFileName )

// Try to open a Z compressed
//  [1/29/2024 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
	// Try to use the compressed file
	std::ifstream in( csFileName, std::ios::binary | std::ios::ate );
	if ( in.good() ) {
		std::istringstream iss;

		//std::cerr << "*-* CDIB::OpenCompressedBMP filename='" << csFileName << std::endl;
		size_t file_size = in.tellg();
		in.seekg( 0, std::ios::beg );
		Bytef* compdata = new Bytef[file_size];
		in.read( (char*)compdata, file_size );
		in.close();

		size_t complen = file_size * 4;	// initial guess
		Bytef* uncompdata = new Bytef[complen];
		while ( Z_OK != uncompress( uncompdata, &complen, compdata, file_size ) ) {
			delete[] uncompdata;
			complen *= 2;
			uncompdata = new Bytef[complen];
		}
		delete[] compdata;

		// Now "complen" has the uncompressed size
		iss.str( std::string( (char*)uncompdata, complen ) );
		delete[] uncompdata;

		return ReadBMP( iss );
	}

	return false;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

int CDIB::CountColors()

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
	assert(GetBitCount()==8);
	unsigned char colors[256],*ptr;
	int nNum=0,i,j,w,d;
	w = Width();
	d = Height();
	memset(colors,0,256);
	for(i=0; i < d; i++) {
		ptr = GetLinePtr(i);
		for(j=0; j < w; j++,ptr++) {
			if(!colors[*ptr]) {
				colors[*ptr]=1;
				nNum++;
			}
		}
	}
	return nNum;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

int CDIB::EnumColors( unsigned char* array)

// [5/26/2020 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
	assert(GetBitCount()==8);
	unsigned char* ptr;
	int nNum=0,i,j,w,d;
	w = Width();
	d = Height();
	memset(array,0,256);
	for(i=0; i < d; i++) {
		ptr = GetLinePtr(i);
		for(j=0; j < w; j++,ptr++) {
			if(!array[*ptr]) {
				array[*ptr]=1;
				nNum++;
			}
		}
	}
	return nNum;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

COLORREF CDIB::PaletteColor(int nIndex)

// [5/26/2020 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
	assert(nIndex < 256);
	RGBQUAD *pRGB= m_pRGB+nIndex;
	return MAKE_RGB(pRGB->rgbRed,pRGB->rgbGreen,pRGB->rgbBlue);
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

bool CDIB::SwitchFrom24(CDIB& dib)

// [5/26/2020 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
	int i,j,w,h,c;
	unsigned char* sPtr,*dPtr;
	unsigned char* index_ptr=NULL;
	RGBQUAD rgb;
	w = Width();
	h = Height();
	index_ptr = (unsigned char*)malloc(0x7FFF+1);
	if(!index_ptr) return false
		;
	memset(CachePtr,0,sizeof(CachePtr));
	for(i=0; i <= 0x7FFF; i++) {
		rgb.rgbRed = (((i & 0x7C00)>>10) << 3) | 0x07;
		rgb.rgbGreen = (((i & 0x3e0)>>5) << 3) | 0x07;
		rgb.rgbBlue = ((i & 0x1F)<<3) | 0x07;
		index_ptr[i] = ClosestColor(&rgb);
	}

	for(i=0; i < h; i++) {
		dPtr = GetLinePtr(i);
		sPtr = dib.GetLinePtr(i);
		for(j=0 ; j < w; j++,dPtr++,sPtr+=3) {
			c = (*sPtr >> 3) | ((*(sPtr+1) >> 3) << 5) | ((*(sPtr+2) >> 3) << 10);
			*dPtr = index_ptr[c];
		}
	}
	free(index_ptr);
	return true;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

void CDIB::FlipHorizontally()

// [5/26/2020 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
  COLORREF Color;
  CDIB     ImageCopy;

  ImageCopy = *this;

  for( int y = 0; y < Height(); ++y ) {
    for( int x = 0; x < Width(); ++x ) {
      Color = ImageCopy.GetPixel( x, y);
      SetPixel( x, Height() - y - 1, Color );
    }
  }
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

void CDIB::FlipVertically()

// [5/26/2020 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
  COLORREF Color;
  CDIB     ImageCopy;

  ImageCopy = *this;

  for( int x = 0; x < Width(); ++x ) {
    for( int y = 0; y < Height(); ++y ) {
      Color = ImageCopy.GetPixel( x, y);
      SetPixel( Width() - x - 1, y, Color );
    }
  }
}
// vim:ts=2
