////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// F A R M     -  Flair advanced rendering module
//
// Copyright European Organization for Nuclear Research( CERN )
// All rights reserved
//
// Authors: Christian.Theis@cern.ch
//			    Vasilis.Vlachoudis@cern.ch
//
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

#ifndef LIGHTMAP_H
#define LIGHTMAP_H

#include "graph/color.h"
#include "math/randsampler.h"
#include "math/vector.h"
#include "math/matrix3.h"
#include "math/random.h"
#include "geometry/ray.h"
#include "geometry/material.h"
#include "geometry/engine.h"
#include "geometry/geometry.h"
#include "dib.h"
#include "bxdfbase.h"

class CLightMap
{
public:
  CLightMap();
  virtual ~CLightMap(void);

	virtual bool    LoadMap( const std::string& FileName, Geometry* pGeometry );
	bool            IsLoaded() const { return (m_pLightMap != nullptr); };
	void            UnLoad() { m_pGeometry->unloadBitmap( m_pLightMap); };

  // define which axis in 3D space will be the horizontal one (U) and which one the vertical (V)
  void            SetUVProjection( const Vector& UAxis, const Vector& VAxis );
  const Matrix3&  GetUVProjection() const { return m_Projection; };
        Matrix3&  GetUVProjection() { return m_Projection; };

  // sample the color received at the intersection point of the incoming ray & return the corresponding light ray
  Color           SampleLightMap( GeometryEngine* pEngine, int SampleNr, const Ray& Out, const Vector& Normal, Ray& LightRay, float& Transmission );

  // check if the ray fired in a certain direction is blocked & return the transmission (1 = unblocked, 0 = blocked, (0,1) transparent object)
  float           IsBlocked( GeometryEngine* pEngine, Ray& LightRay );

  // retrieve light emission of a ray going into a specific direction
  virtual Color   LightEmission( const Vector& Dir );

  virtual int     GetIntegralPatchSize() const { return m_IntegralStep; };
  virtual void    SetIntegralPatchSize( unsigned int Val ) { m_IntegralStep = Val; };

  virtual double  GetArea() { return 4.0 * m_Extension * m_Extension * PI; };

  // total distance of the lightmap surrounding the scene
  void            SetDistance( double Val ) { m_Extension = Val; }
  double          GetDistance() const { return m_Extension; };

  // intensity
	void            SetIntensity( float Val ) { m_Intensity = std::max( 0.0f, Val ); };
	float           GetIntensity() const { return m_Intensity; }

  // attenuation
  bool            Attenuate() const { return m_bAttenuate; };
  void            Attenuate( bool bFlag ) { m_bAttenuate = bFlag; };
  double          AttenuationFactor() { return 1.0 / (PI * m_Extension * m_Extension); };

  void            EnableShadows( bool bFlag ) { m_bShadows = bFlag;  }
  bool            AreShadowsEnabled() const { return m_bShadows; }

protected:

  // sample the color + the direction of the light ray
  // this function should be overloaded if the sampling is to be changed
  virtual Color   Sample_f( GeometryEngine* pEngine, const Ray& Out, const Vector& Normal, int SampleNr, Vector& LightDir );

private:
  Color           LookUpColor( const Vector& Dir );

protected:
  CDIB*           m_pLightMap;
  unsigned int    m_IntegralStep;
  double          m_Extension;
  float           m_Intensity;
  bool            m_bAttenuate;
  bool            m_bShadows;

  Matrix3         m_Projection;
  Matrix3         m_InvProjection;
  Geometry*       m_pGeometry;
};
#endif
// vim:ts=2
