////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// F A R M     -  Flair advanced rendering module
//
// Copyright European Organization for Nuclear Research( CERN )
// All rights reserved
//
// Authors: Christian.Theis@cern.ch
//			    Vasilis.Vlachoudis@cern.ch
//
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

#include "matorennayar.h"

CMatOrenNayar::CMatOrenNayar( CBaseIntegrator* pInt, double Sigma /* = 15 */, bool bNoiseTex /*= false*/, bool bDisplacement /*= false*/, double BumpScaling /*= 1*/ ) : CPhysMatInterface( pInt, CBxDFBase::BSDF_DIFFUSE, false /* false = no MC sampling required */  ),
m_bUseNoise( bNoiseTex ), m_bDisplacement( bDisplacement )
{
	m_BxDF.SetSigma( Sigma );

	if ( !bDisplacement ) { // bump mapping only
		m_NoiseTexture.SetMinMax( 0.7, 1.0 );       // restrict noise
		m_NoiseTexture.m_Style = CNoiseTexture::cTurbulent2;
	}
	else {
		m_NoiseTexture.SetMinMax( 0.0, 0.7 );       // restrict noise
		m_NoiseTexture.m_Style = CNoiseTexture::cTurbulent;
	}

	m_BumpScaling = BumpScaling;
}

CMatOrenNayar::~CMatOrenNayar(void)
{
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

const Vector& CMatOrenNayar::GetShadingNormal( const Ray* pCurRay )

//  [9/25/2023 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
	// check if we need to evaluate and update our cached shading normal based on Perlin noise bump-mapping

	if ( (m_bUseNoise || m_bDisplacement) && (!m_pCachedShadingRay || pCurRay != m_pCachedShadingRay) ) {
		m_CachedShadingNormal = m_NoiseTexture.GetBumpNormal( pCurRay, m_BumpScaling );    // TODO: replace noise-scaling with scaling parameter from hit material
		m_pCachedShadingRay = const_cast<Ray*>(pCurRay);
	}
	else if ( !(m_bUseNoise && m_bDisplacement) ) {
		return pCurRay->normal;
	}

	return m_CachedShadingNormal;

}


////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

void CMatOrenNayar::ModifyShadingGeometry( Ray* pCurRay, GeometryEngine* pEngine )

//  implement displacement mapping if procedural noise textures are activated
//
//  [9/25/2023 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
	if ( m_bDisplacement )
		m_NoiseTexture.DisplaceHitRay( pCurRay, pEngine, m_BumpScaling);
}


////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// Oren-Nayar with textures and displacement mapping
//
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////


////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

const Vector& CMatOrenNayarTex::GetShadingNormal( const Ray* pCurRay )

// Oren Nayar with normal displacement texture maps
// [10/20/2023 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
	return GetBumpNormalFromTex( pCurRay );
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

void CMatOrenNayarTex::ModifyShadingGeometry( Ray* pCurRay, GeometryEngine* pEngine )

// Oren Nayar with normal displacement texture maps
// [10/20/2023 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
	ModifyShadingGeometryByTex( pCurRay, pEngine );
}


////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

Color CMatOrenNayarTex::GetHitColor( const Ray* pRay )

// Override hit color with the one from the image texture
//  [11/1/2023 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
	if( !m_ImageTex.IsTextureLoaded() )
		return pRay->hitRegion()->color();

	return m_ImageTex.GetTexColor( pRay );
}

// vim:ts=2