////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// F A R M     -  Flair advanced rendering module
//
// Copyright European Organization for Nuclear Research( CERN )
// All rights reserved
//
// Authors: Christian.Theis@cern.ch
//			    Vasilis.Vlachoudis@cern.ch
//
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

#ifndef MAT_OREN_NAYAR_H
#define MAT_OREN_NAYAR_H

#include <string>

#include "physmatinterface.h"
#include "bxdfbase.h"
#include "bxdforennayar.h"
#include "texture.h"
#include "noisetexture.h"



class CMatOrenNayar : public CPhysMatInterface
{
public:
  CMatOrenNayar( CBaseIntegrator* pInt, double Sigma = 15, bool bNoiseTex = false, bool bDisplacement = false, double BumpScaling = 1 );
  virtual ~CMatOrenNayar(void);

	// f( Out, In) evaluates the BxDF for the in & outgoing pair of rays
	// Out = ray coming FROM the camera, In = ray going towards the light
	virtual Color f( const Ray& Out, const Ray& In, const Vector& Normal, const Material& RenderMat, const Color& LightColor, double Power, float Shade = 1.0f ) override
	{
		return m_BxDF.f( Out, In, Normal, RenderMat, LightColor, Power, Shade );
	}

	// By default only diffuse sampling is performed. This function needs to be overwritten by the actual physical materials that
	// have reflective and probably translucent components!
	virtual Color Sample_f( const Ray& Out, Ray& In, float& Pdf, const Vector& Normal, const Material& RenderMat, double* pRandNr, CBxDFBase::BxDFType Type, const Color& LightColor = Color( 1.0, 1.0, 1.0 ), double Power = 1.0, float Shade = 1.0f ) override
	{
		return m_BxDF.Sample_f( Out, In, Pdf, Normal, RenderMat, pRandNr, Type, LightColor, Power, Shade );
	}

	// override shading normal & geometry modififiers
	virtual const Vector& GetShadingNormal( const Ray* pCurRay ) override;
	virtual void ModifyShadingGeometry( Ray* pCurRay, GeometryEngine* pEngine ) override;

protected:
  CBxDFOrenNayar	m_BxDF;

	bool						m_bUseNoise;
	bool						m_bDisplacement;

};

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// oren nayar with normal displacement maps
//
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
class CMatOrenNayarTex : public CMatOrenNayar
{
public:
	CMatOrenNayarTex( CBaseIntegrator* pInt, const std::string& TexFileName, double TexDimU, double TexDimV, double Sigma = 15, double BumpScaling = 1 ) : CMatOrenNayar( pInt, Sigma, true, true, BumpScaling )
	{
		m_TexFileName = TexFileName;
		m_TexDimU = TexDimU;
		m_TexDimV = TexDimV;
	};

	virtual ~CMatOrenNayarTex() = default;

	// override shading normal & geometry modififiers
	virtual const Vector& GetShadingNormal( const Ray* pCurRay ) override;
	virtual void ModifyShadingGeometry( Ray* pCurRay, GeometryEngine* pEngine ) override;

	// override color with the one from the image texture
	Color GetHitColor( const Ray* pRay ) override;

};

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// Soil with normal displacement maps
//
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

class CMatSoilTex : public CMatOrenNayarTex
{
public:
	CMatSoilTex( CBaseIntegrator* pInt, const std::string& TexFileName, double TexDimU, double TexDimV, double Sigma = 15, double BumpScaling = 1 ) : CMatOrenNayarTex( pInt, TexFileName, TexDimU, TexDimV, Sigma, BumpScaling )
	{
	}

	virtual ~CMatSoilTex() = default;

	// override color with the one from the image texture
	Color GetHitColor( const Ray* pRay ) override
	{
		return MixBaseWithTexColor( pRay, 0.02 );
	}

};


#endif
// vim:ts=2