////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// F A R M     -  Flair advanced rendering module
//
// Copyright European Organization for Nuclear Research( CERN )
// All rights reserved
//
// Authors: Christian.Theis@cern.ch
//			    Vasilis.Vlachoudis@cern.ch
//
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////


#include "matplastic.h"

CMatPlastic::CMatPlastic( CBaseIntegrator* pInt, const float cRoughness /* = 1.0 */, bool bNoiseTex /*= false*/) : CPhysMatInterface( pInt, static_cast<CBxDFBase::BxDFType>(CBxDFBase::BSDF_DIFFUSE | CBxDFBase::BSDF_SPECULAR | CBxDFBase::BSDF_REFLECTION | CBxDFBase::BSDF_TRANSMISSION), true /* true = MC sampling required for the microfacet distributions for example */ ),
                                                                                                                                                 m_bUseNoise( bNoiseTex )
{
  // roughness higher values indicate a higher roughness -> sharp specular highlight, lower values -> broad specular highlight

  m_pBxDFDiffuse    = new CBxDFPhong;
  m_pFresnelDiel    = new CFresnelPreDefDiel( 1.0, 1.47 );
  m_pBlinnMicroDist = new CBlinnMicroDist( cRoughness );
  m_pBxDFSpecular   = new CBxDFMicrofacet( m_pFresnelDiel, m_pBlinnMicroDist );
  m_pBxDFSpecReflection = new CBxDFSpecReflection( m_pFresnelDiel );
  m_pBxDFSpecTransmission = new CBxDFSpecTransmission;

	m_NoiseTexture.SetMinMax( 0.0, 1.0 );       // restrict noise
	m_NoiseTexture.m_Style = CNoiseTexture::cArtistic;

}

CMatPlastic::~CMatPlastic(void)
{
  delete m_pFresnelDiel;
  delete m_pBlinnMicroDist;
  delete m_pBxDFDiffuse;
  delete m_pBxDFSpecular;
  delete m_pBxDFSpecReflection;
  delete m_pBxDFSpecTransmission;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////

Color CMatPlastic::f( const Ray& Out, const Ray& In, const Vector& Normal, const Material& RenderMat, const Color& LightColor, double Power, float Shade /*= 1.0*/ )

// f( Out, In) evaluates the BxDF for the in & outgoing pair of rays
////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
  UNUSED( Power );

  Color Diffuse, Specular;

  if( m_pBlinnMicroDist)
    m_pBlinnMicroDist->SetExp( std::max( 1.0, RenderMat.shine() ) );

	double NoiseScaling = Out.hitRegion()->region()->material()->roughness();
  double NoiseFact = m_bUseNoise ? m_NoiseTexture.Value( 0, 0, Out.hit(), NoiseScaling )[0] : 1.0;    // for the moment we only use the 1st channel of our noise, though we could use 3 of them if we want to influence

  Diffuse   = m_pBxDFDiffuse->f( Out, In, Normal, RenderMat, LightColor, Shade );
  Specular  = m_pBxDFSpecular->f( Out, In, Normal, RenderMat, LightColor, Shade );

  //return NoiseFact * (Diffuse * (1.0 - RenderMat.specular() ) + Specular * RenderMat.specular());
  return (Diffuse * NoiseFact * (1.0 - RenderMat.specular()) + Specular * RenderMat.specular());
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////

Color CMatPlastic::Sample_f( const Ray& Out, Ray& In, float& Pdf, const Vector& Normal, const Material& RenderMat, double* pRandNr, CBxDFBase::BxDFType Type, const Color& LightColor /*= Color( 1.0, 1.0, 1.0 )*/, double Power /*= 1.0*/, float Shade /*= 1.0*/ )

// Sample_f samples an incoming ray matching to the outgoing ray + its Pdf. Furthermore it evaluates the
// BxDF according to In & Out and returns the color. This is used for reflection, refraction etc. which will
// continue recursively.
////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
  UNUSED( Power );

  if( Type == CBxDFBase::BSDF_REFLECTION )
    return m_pBxDFSpecReflection->Sample_f( Out, In, Pdf, Normal, RenderMat, pRandNr, Type, LightColor, Shade );
  else if( Type == CBxDFBase::BSDF_TRANSMISSION )
    return m_pBxDFSpecTransmission->Sample_f( Out, In, Pdf, Normal, RenderMat, pRandNr, Type, LightColor, Shade );

  return CPhysMatInterface::Sample_f( Out, In, Pdf, Normal, RenderMat, pRandNr, Type, LightColor, Shade );
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// Rough plastic with microfacet distributions and normal displacements
//
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

CMatPlasticEx::CMatPlasticEx( CBaseIntegrator* pInt, CMicrofacetDistribution* pMicroDist, bool bNoiseTex /*= false*/, bool bDisplacement /*= false */, double BumpScaling /* = 1*/) : CMatPlastic(pInt, 0, bNoiseTex),
m_bDisplacement( bDisplacement ), m_BumpScaling( BumpScaling), m_pMicroFacets( pMicroDist )
{
	m_pBxDFDiffuse = new CBxDFLambert;
	m_pFresnelDiel = new CFresnelPreDefDiel( 1.0, 1.47 );
	m_pBlinnMicroDist = nullptr;
	m_pBxDFSpecular = new CBxDFMicrofacet( m_pFresnelDiel, m_pMicroFacets);
	m_pBxDFSpecReflection = new CBxDFSpecReflection( m_pFresnelDiel );
	m_pBxDFSpecTransmission = new CBxDFSpecTransmission;

  if ( m_bDisplacement ) {
    m_NoiseTexture.SetMinMax( 0.3, 0.8 );
    m_NoiseTexture.m_Style = CNoiseTexture::cTurbulent2;
  }
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

const Vector& CMatPlasticEx::GetShadingNormal( const Ray* pCurRay )

//  [9/12/2023 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
	// check if we need to evaluate and update our cached shading normal based on Perlin noise bump-mapping

	if ( (m_bUseNoise || m_bDisplacement) && (!m_pCachedShadingRay || pCurRay != m_pCachedShadingRay) ) {
		m_CachedShadingNormal = m_NoiseTexture.GetBumpNormal( pCurRay, m_BumpScaling );
		m_pCachedShadingRay = const_cast<Ray*>(pCurRay);
	}
	else if ( !(m_bUseNoise && m_bDisplacement) ) {
		return pCurRay->normal;
	}

	return m_CachedShadingNormal;

}


////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

void CMatPlasticEx::ModifyShadingGeometry( Ray* pCurRay, GeometryEngine* pEngine )

//  implement displacement mapping if procedural noise textures are activated
//
//  [9/14/2023 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
	if ( m_bDisplacement )
		m_NoiseTexture.DisplaceHitRay( pCurRay, pEngine, m_BumpScaling);
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// rough plastic with textures & displacement mapping
//
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

CMatRoughPlasticTex::CMatRoughPlasticTex( CBaseIntegrator* pInt, double Roughness, const std::string& TexFileName, double TexDimU, double TexDimV, double BumpScaling /*= 1*/ ) :
  CMatPlasticEx( pInt, new CTrowbridgeReitzDist( CTrowbridgeReitzDist::RoughnessToAlpha( Roughness ), CTrowbridgeReitzDist::RoughnessToAlpha( 2 * Roughness ) ), false, true, BumpScaling )
{
	m_TexFileName = TexFileName;
	m_TexDimU = TexDimU;
	m_TexDimV = TexDimV;

  //InitTextures();
}

// vim:ts=2