////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// F A R M     -  Flair advanced rendering module
//
// Copyright European Organization for Nuclear Research( CERN )
// All rights reserved
//
// Authors: Christian.Theis@cern.ch
//			    Vasilis.Vlachoudis@cern.ch
//
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

#include "noisetexture.h"

using namespace std;


////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

Vector CNoiseTexture::Value( const double u, const double v, const Vector& p, double Scale ) const

//  [9/5/2023 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
	UNUSED( u );
	UNUSED( v );

	switch ( m_Style ) {
		case Style::cNormal:
			// return Vector( m_Min, m_Min, m_Min ) + (Vector( 0.5, 0.5, 0.5 ) * (1 + m_NoiseGen.Noise( m_Scale * p ))) * (m_Max - m_Min);    // Non-turbulent
			// return Vector( m_Min, m_Min, m_Min ) + (Vector( (m_Max - m_Min) * 0.5, (m_Max - m_Min) * 0.5, (m_Max - m_Min) * 0.5 ) * (m_NoiseGen.Noise( m_Scale * p )));    // Non-turbulent
			return Vector( m_Min, m_Min, m_Min ) + Vector( m_NoiseGen.Turbulence( Scale * p, 2), m_NoiseGen.Turbulence( Scale * p, 2 ), m_NoiseGen.Turbulence( Scale * p, 2) ) * (m_Max - m_Min);
		case Style::cTurbulent:
			return Vector( m_Min, m_Min, m_Min ) + Vector( m_NoiseGen.Turbulence( Scale * p ), m_NoiseGen.Turbulence( Scale * p ), m_NoiseGen.Turbulence( Scale * p ) ) * (m_Max - m_Min);
		case Style::cMarble:
			//return Vector( m_Min, m_Min, m_Min ) + (Vector( 0.5, 0.5, 0.5 ) * (1 + std::sin( (m_Scale * p.z) + (10 * m_NoiseGen.Turbulence( ( p )) )))) * (m_Max - m_Min);
			return Vector( m_Min, m_Min, m_Min ) + (Vector( 0.5, 0.5, 0.5 ) * (1 + std::sin( (Scale * p.y) + (10 * m_NoiseGen.Turbulence( ( p )) )))) * (m_Max - m_Min);
		case Style::cArtistic:
			//return Vector( m_Min, m_Min, m_Min ) + (Vector( 0.5, 0.5, 0.5 ) * (1 + std::sin( (m_Scale * p.x) + (5 * m_NoiseGen.Turbulence( ( m_Scale * p )) )))) * (m_Max - m_Min);
			return Vector( m_Min, m_Min, m_Min ) + (Vector( 0.5, 0.5, 0.5 ) * (1 + std::sin( (Scale * p.y) + (5 * m_NoiseGen.Turbulence( (Scale * p) )) ))) * (m_Max - m_Min);
		case Style::cTurbulent2:
			return Vector( m_Min, m_Min, m_Min ) + (Vector( 0.5, 0.5, 0.5 ) * (1 + std::sin( (Scale * p.x) + (3 * m_NoiseGen.Turbulence( (p), 5 )) ))) * (m_Max - m_Min);
	}

	return Vector( m_Min, m_Min, m_Min ) + (Vector( 0.5, 0.5, 0.5 ) * (1 + m_NoiseGen.Noise( Scale * p ))) * (m_Max - m_Min);    // Non-turbulent
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

Vector CNoiseTexture::GetBumpNormal( const Ray* pCurRay, double BumpScaling )

//  [9/12/2023 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
	constexpr double cEps = 0.01;

	double TexScaling = pCurRay->hitRegion()->region()->material()->roughness();

	double F0 = Value( 0, 0, pCurRay->hit(), TexScaling )[0];
	double FX = Value( 0, 0, Vector( pCurRay->hit().x + cEps, pCurRay->hit().y, pCurRay->hit().z ), TexScaling )[0];
	double FY = Value( 0, 0, Vector( pCurRay->hit().x, pCurRay->hit().y + cEps, pCurRay->hit().z ), TexScaling )[0];
	double FZ = Value( 0, 0, Vector( pCurRay->hit().x, pCurRay->hit().y, pCurRay->hit().z + cEps ), TexScaling )[0];

	Vector dF( (FX - F0) / cEps, (FY - F0) / cEps, (FZ - F0) / cEps );   // calculate approximated gradient based on noise
	Vector RetVec = (pCurRay->normal - (dF * BumpScaling));
	RetVec.normalize();
	return RetVec;
}


////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

void CNoiseTexture::DisplaceHitRay( Ray* pCurRay, GeometryEngine* pEngine, double BumpScaling )

// displace the hit point of the hit ray by the texture to implement displacement mapping
//
//  [1/9/2024 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
	constexpr double cEps = 0.01;

 	// flag required cache update as we change the last ray direction
 	pEngine->incBodyCheckId();

	double TexScaling = pCurRay->hitRegion()->region()->material()->roughness();

	double F0 = Value( 0, 0, pCurRay->hit(), TexScaling )[0];
	double FX = Value( 0, 0, Vector( pCurRay->hit().x + cEps, pCurRay->hit().y, pCurRay->hit().z ), TexScaling )[0];
	double FY = Value( 0, 0, Vector( pCurRay->hit().x, pCurRay->hit().y + cEps, pCurRay->hit().z ), TexScaling )[0];
	double FZ = Value( 0, 0, Vector( pCurRay->hit().x, pCurRay->hit().y, pCurRay->hit().z + cEps ), TexScaling )[0];

	Vector dF( (FX - F0) / cEps * pCurRay->normal.x, (FY - F0) / cEps * pCurRay->normal.y, (FZ - F0) / cEps * pCurRay->normal.z );   // N^ * dF

	dF.normalize();

	// NOTE: This is not true displacement as this would either require triangulated mesh displacement of the geometry or
	//			 iterative re-tracking, which is too expensive. However, visually it gives a satisfying result if we are not too close to notice

	Vector CurHit = pCurRay->segment().hit( -SMALL3D4 );
	CurHit += dF * BumpScaling;		// displaced position

	Vector dir = CurHit - pCurRay->segment().pos;
	dir.normalize();

	pCurRay->segment().dir = dir;

}
