////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// F A R M     -  Flair advanced rendering module
//
// Copyright European Organization for Nuclear Research( CERN )
// All rights reserved
//
// Authors: Christian.Theis@cern.ch
//			    Vasilis.Vlachoudis@cern.ch
//
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

#pragma once
#ifndef TEXTURE_H
#define TEXTURE_H

#include "math/vector.h"
#include "math/matrix3.h"
#include "math/bmath.h"
#include "dib.h"
#include "bxdfbase.h"
#include "geometry/gbody.h"
#include "geometry/spherebody.h"
#include "geometry/cylbody.h"
#include "geometry/engine.h"
#include "geometry/geometry.h"


class CTextureBase
{
public:
	virtual ~CTextureBase() = default;

	// this must be overloaded by the respective texture class which implements its own mapping
	virtual bool GetTextureCoordinates( const Ray* pCurRay, double& u, double& v ) const = 0;

protected:

	enum Axis { cXAxis, cYAxis, cZAxis };

	bool GetSphericalTexCoords( const Ray* pCurRay, double& u, double& v ) const
	{

		double Radius = ((const GSPHBody*)pCurRay->hitBody()->body())->R();

		// we must use the hit point in local coordinates for the UV mapping!
		Vector ObjHitPoint = pCurRay->hitInObjCoords();

		auto Theta = acos( ObjHitPoint.z / Radius );
		auto Phi = atan2( ObjHitPoint.y / Radius, ObjHitPoint.x / Radius);

		constexpr double cOneOver2Pi = 1.0 / (2 * M_PI);
		constexpr double cOneOverPi = 1.0 / M_PI;
		constexpr double cTwoPI = 2 * M_PI;

		if ( Phi < 0.0 )
			Phi += cTwoPI;

		u = Phi * cOneOver2Pi;
		v = (M_PI - Theta) * cOneOverPi;


// 		assert( !isnan( u ) );
// 		assert( !isnan( v ) );

		return true;
	}

	bool GetCylinderTexCoords( const Ray* pCurRay, Axis Direction, double& u, double& v ) const
	{
		double RadialComponent = 0;
		double HeightComponent = 0;

		double Radius = ((const GRCCBody*)pCurRay->hitBody()->body())->R();

		// we must use the hit point in local coordinates for the UV mapping!
		Vector ObjHitPoint = pCurRay->hitInObjCoords();

		if ( Direction == cZAxis ) {
			RadialComponent = ObjHitPoint.x;
			HeightComponent = ObjHitPoint.z;
		}
		else if ( Direction == cYAxis ) {
			RadialComponent = ObjHitPoint.y;
			HeightComponent = ObjHitPoint.x;
		}
		else {
			RadialComponent = ObjHitPoint.x;
			HeightComponent = ObjHitPoint.y;
		}

		auto Theta = acos( RadialComponent / Radius );

		u = Theta;
		v = HeightComponent;

// 		assert( !isnan( u ) );
// 		assert( !isnan( v ) );

		return true;
	}

	// get planar tex coordinates based on the dominant axis of the normal
	bool GetPlanarTexCoords( const Ray* pCurRay, double& u, double& v ) const
	{

		// we must use the hit point in local coordinates for the UV mapping!
		Vector ObjHitPoint = pCurRay->hitInObjCoords();

		// via the dominant component of the normal we can identify the tangential space
		// within which we map the UV coordinates of the texture
		int DominantComponent = pCurRay->normal.dominant();

		if ( DominantComponent == 0 ) {
			u = ObjHitPoint.y;
			v = ObjHitPoint.z;
		}
		else if ( DominantComponent == 1 ) {
			u = ObjHitPoint.z;
			v = ObjHitPoint.x;
		}
		else {
			u = ObjHitPoint.x;
			v = ObjHitPoint.y;
		}

		return true;
	}

	// get planar tex coordinates based on the passed normal direction
	bool GetPlanarTexCoords( const Ray* pCurRay, Axis Direction, double& u, double& v ) const
	{
		// we must use the hit point in local coordinates for the UV mapping!
		Vector ObjHitPoint = pCurRay->hitInObjCoords();

		if ( Direction == cXAxis ) {
			u = ObjHitPoint.y;
			v = ObjHitPoint.z;
		}
		else if ( Direction == cYAxis ) {
			u = ObjHitPoint.z;
			v = ObjHitPoint.x;
		}
		else {
			u = ObjHitPoint.x;
			v = ObjHitPoint.y;
		}

		return true;
	}


	bool GetBoxTexCoords( const Ray* pCurRay, double& u, double& v ) const
	{

		// we must use the hit point in local coordinates for the UV mapping!
		Vector ObjHitPoint = pCurRay->hitInObjCoords();

		// via the dominant component of the normal we can identify the tangential space
		// within which we map the UV coordinates of the texture
		int DominantComponent = pCurRay->normal.dominant();
		if ( DominantComponent == 0 ) {
			u = ObjHitPoint.x;
			v = ObjHitPoint.z;
		}
		else if ( DominantComponent == 1 ) {
			u = ObjHitPoint.y;
			v = ObjHitPoint.x;
		}
		else {
			u = ObjHitPoint.z;
			v = ObjHitPoint.y;
		}

		return true;
	}

	bool GetPositionalSphericalTexCoords( const Ray* pCurRay, double& u, double& v ) const
	{
		// positional spherical mapping by drawing vectors from the object's center through the hitpoint onto a virtual surrounding sphere
		Vector HitPoint = pCurRay->hit();
		auto pHitBody = pCurRay->hitBody()->body();
		auto Center = pHitBody->bbox().centroid();

		Vector TexVec = HitPoint - Center;
		TexVec.normalize();

		u = asin( TexVec.x ) / M_PI;
		v = asin( TexVec.y ) / M_PI;

		return true;
	}

	// return texture coordinates based on the topology of the object that was hit
	bool GetGenericTexCoords( const Ray* pCurRay, double& u, double& v ) const
	{
		// try to map texture coordinates based on hit body type
		if ( pCurRay->hitBody() ) {
			auto HitType = pCurRay->hitBody()->body()->type();

			// try to do correct mapping via the topology of the hit primitive and fall back
			// to planar mapping otherwise

			// spheres
			if ( HitType >= GBody::Type::SPH && HitType <= GBody::Type::ELL ) {
				return GetSphericalTexCoords( pCurRay, u, v );
			}

			// axis aligned infinite cylinders and cones
			else if ( HitType == GBody::Type::C_X || HitType == GBody::Type::CX || HitType == GBody::Type::XCC || HitType == GBody::Type::XEC ||
								HitType == GBody::Type::K_X || HitType == GBody::Type::KX ) {
				return GetCylinderTexCoords( pCurRay, Axis::cXAxis, u, v );
			}

			else if ( HitType == GBody::Type::C_Y || HitType == GBody::Type::CY || HitType == GBody::Type::YCC || HitType == GBody::Type::YEC ||
								HitType == GBody::Type::K_Y || HitType == GBody::Type::KY ) {
				return GetCylinderTexCoords( pCurRay, Axis::cYAxis, u, v );
			}
			else if ( HitType == GBody::Type::C_Z || HitType == GBody::Type::CZ || HitType == GBody::Type::ZCC || HitType == GBody::Type::ZEC ||
								HitType == GBody::Type::K_Z || HitType == GBody::Type::KZ ) {
				return GetCylinderTexCoords( pCurRay, Axis::cZAxis, u, v );
			}

			// arbitrary finite cylinders
			else if ( HitType == GBody::Type::RCC || HitType == GBody::Type::REC ) {
				if( fabs( pCurRay->hitBody()->body()->vectorZ().x ) < SMALL3D3 && fabs( pCurRay->hitBody()->body()->vectorZ().y ) < SMALL3D3 )
					return GetCylinderTexCoords( pCurRay, Axis::cZAxis, u, v );
				else if( fabs( pCurRay->hitBody()->body()->vectorZ().x ) < SMALL3D3 && fabs( pCurRay->hitBody()->body()->vectorZ().z ) < SMALL3D3 )
					return GetCylinderTexCoords( pCurRay, Axis::cYAxis, u, v );
				else if ( fabs( pCurRay->hitBody()->body()->vectorZ().y ) < SMALL3D3 && fabs( pCurRay->hitBody()->body()->vectorZ().z ) < SMALL3D3 )
					return GetCylinderTexCoords( pCurRay, Axis::cZAxis, u, v );	// ATTENTION: Z-axis mapping this is done on purpose to ensure correct mapping on the cylindrical surface instead of the base
				else
					return GetBoxTexCoords( pCurRay, u, v );
			}

			// arbitrary infinite planes
			else if ( HitType == GBody::Type::PLA ) {
				if ( fabs( pCurRay->hitBody()->body()->vectorZ().x ) < SMALL3D3 && fabs( pCurRay->hitBody()->body()->vectorZ().y ) < SMALL3D3 )
					return GetPlanarTexCoords( pCurRay, Axis::cZAxis, u, v );
				else if ( fabs( pCurRay->hitBody()->body()->vectorZ().x ) < SMALL3D3 && fabs( pCurRay->hitBody()->body()->vectorZ().z ) < SMALL3D3 )
					return GetPlanarTexCoords( pCurRay, Axis::cXAxis, u, v );
				else if ( fabs( pCurRay->hitBody()->body()->vectorZ().y ) < SMALL3D3 && fabs( pCurRay->hitBody()->body()->vectorZ().z ) < SMALL3D3 )
					return GetPlanarTexCoords( pCurRay, Axis::cZAxis, u, v );
				else
					return GetBoxTexCoords( pCurRay, u, v );
			}


			// boxes, planes and all the rest
			else if ( HitType == GBody::Type::BOX || (HitType == GBody::Type::YZP || HitType == GBody::Type::PX ) )
				return GetBoxTexCoords( pCurRay, u, v );
			else if ( HitType == GBody::Type::RPP || HitType == GBody::Type::QUA || HitType == GBody::Type::XYP || HitType == GBody::Type::PZ  )
				return GetPlanarTexCoords( pCurRay, u, v );
			else if ( (HitType == GBody::Type::XZP || HitType == GBody::Type::PY) ) {
				return GetPlanarTexCoords( pCurRay, Axis::cXAxis, u, v ); // ATTENTION: X-axis mapping is done on purpose to ensure correct mapping
			}
			else
				return GetPlanarTexCoords( pCurRay, u, v );
		}

		// TODO: fix this situation
		//std::cerr << "Tex coordinate generation: WARNING: no hitbody() but hitzone() = " << pCurRay->hitZone()->name() << std::endl;
		u = 0; v = 0;
		return false;
	}
};

class CDisplacementTexture
{
public:
	virtual ~CDisplacementTexture() = default;

		// return displacement vector for a specific position
	virtual Vector  GetBumpNormal( const Ray* pCurRay, double Scale = 1.0 ) = 0;

	// displace the hit ray by the normal mapping
	virtual void		DisplaceHitRay( Ray* pCurRay, GeometryEngine* pEngine, double BumpScaling ) { UNUSED( pCurRay ); UNUSED( BumpScaling ); UNUSED( pEngine );  };
};

class CImageTexture : public CTextureBase
{
public:
	CImageTexture() : m_pTexture(nullptr), m_TexDimU(1), m_TexDimV( 1 ), m_pGeometry(nullptr) {};
	CImageTexture( const std::string& FileName, double DimU = 1, double DimV = 1 );
	virtual ~CImageTexture() = default;

	// this needs to set as the bitmap pool is in the geometry class
	void SetGeometry( Geometry* pGeometry ) { m_pGeometry = pGeometry; };

	//
	bool		LoadTexture( const std::string& FileName );
	void		UnloadTexture() { assert( m_pGeometry ); m_pGeometry->unloadBitmap( m_pTexture ); m_pTexture = nullptr; }
	bool		IsTextureLoaded() const { return (m_pTexture != nullptr); };

	// look up color in a texture for the respective hit point
	virtual Color		GetTexColor( const Ray* pCurRay );

	// set/get the actual dimension of the texture in world units.
	void		GetTexDim( double& U, double& V ) const { U = m_TexDimU; V = m_TexDimV; }
	void		SetTexDim( double U, double V ) { m_TexDimU = U; m_TexDimV = V; }

	// return texture coordinates based on the topology of the object that was hit
	bool GetTextureCoordinates( const Ray* pCurRay, double& u, double& v ) const override
	{
		return GetGenericTexCoords( pCurRay, u, v );
	}

protected:
	CDIB*		m_pTexture;
	double	m_TexDimU;
	double	m_TexDimV;
	Geometry* m_pGeometry;
};

// normal displacement texture base class
class CNormalTex : public CImageTexture, public CDisplacementTexture
{
public:

	CNormalTex() : CImageTexture(), CDisplacementTexture() {};
	CNormalTex( const std::string& FileName ) : CImageTexture( FileName ), CDisplacementTexture() {};
	virtual ~CNormalTex() = default;

	// return displacement vector for a specific position
	virtual Vector GetBumpNormal( const Ray* pCurRay, double Scale = 1.0 ) override;

	// displace the hit ray by the normal mapping
	virtual void		DisplaceHitRay( Ray* pCurRay, GeometryEngine* pEngine, double BumpScaling ) override;

	// this must be overloaded by the respective texture class which implements its own mapping
	virtual bool GetTextureCoordinates( const Ray* pCurRay, double& u, double& v ) const override = 0;

	// get color of the texture at hit point
	Color		GetTexColor( const Ray* pCurRay) override
	{
		Vector Tmp( GetTexValue( pCurRay ) );
		Color  TmpCol( Tmp.x, Tmp.y, Tmp.z );

		return TmpCol;
	}

	// return normal from texture file via texture coordinates
	Vector GetNormalFromTex( const Ray* pCurRay, double u, double v  );

protected:
	Vector	GetTexValue( const Ray* pCurRay );

};

//
// normal displacement texture with spherical mapping
//
class CSphericalNormalTex : public CNormalTex
{
public:
	CSphericalNormalTex() {};
	CSphericalNormalTex( const std::string& FileName ) : CNormalTex( FileName ) {};
	virtual ~CSphericalNormalTex() = default;

	// return spherical texture coordinates
	bool GetTextureCoordinates( const Ray* pCurRay, double& u, double& v ) const override	{
		return GetSphericalTexCoords( pCurRay, u, v );
	}
};

//
// normal displacement texture with cylindrical mapping
//
class CCylNormalTex : public CNormalTex
{
public:
	CCylNormalTex() {};
	CCylNormalTex( const std::string& FileName ) : CNormalTex( FileName ) {};
	virtual ~CCylNormalTex() = default;

	// return spherical texture coordinates
	bool GetTextureCoordinates( const Ray* pCurRay, double& u, double& v ) const override {
		return GetCylinderTexCoords( pCurRay, Axis::cZAxis, u, v );
	}
};

//
// normal displacement texture with planar mapping w.r.t. dominant coordinate
//
class CPlanarNormalTex : public CNormalTex
{
public:
	CPlanarNormalTex() {};
	CPlanarNormalTex( const std::string& FileName ) : CNormalTex( FileName ) {};
	virtual ~CPlanarNormalTex() = default;

	// return spherical texture coordinates
	bool GetTextureCoordinates( const Ray* pCurRay, double& u, double& v ) const override
	{
		return GetPlanarTexCoords( pCurRay, u, v );
	}
};

class CGenericNormalTex : public CNormalTex
{
public:
	CGenericNormalTex() {};
	virtual ~CGenericNormalTex() = default;

	// return texture coordinates based on the topology of the object that was hit
	bool GetTextureCoordinates( const Ray* pCurRay, double& u, double& v ) const override
	{
		return GetGenericTexCoords( pCurRay, u, v );
	}
};
#endif
// vim:ts=2
