/*
 * Copyright European Organization for Nuclear Research (CERN)
 * All rights reserved
 *
 *
 * Author:	Christian.Theis@cern.ch
 *
 * Date: 20-Nov-2020
 */
#ifndef BOUNDING_VOL_HIERARCHY_H
#define BOUNDING_VOL_HIERARCHY_H

#ifndef UNUSED
  #define UNUSED(expr) (void)(expr)
#endif

#include <vector>
#include <memory>
#include <list>

#include "geo.h"
#include "math/bbox.h"
#include "math/vector.h"
#include "geometry/engine.h"
#include "geometry/cbody.h"
#include "math/morton.h"


class Ray;
class VZone;
class CBody;

int QSplit( VZone** ppNodeList, int NrOfNodes, double PivotVal, int Axis);
void MergeBoundingBoxes( const Vector& Box1Min, const Vector& Box1Max, const Vector& Box2Min, const Vector& Box2Max, Vector& NewBoxMin, Vector& NewBoxMax );

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
//                                              Standard mid-point split bounding volume hierarchies
//
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

class CBoundingVolHierarchy
{
public:
  enum  { cXAxis = 0, cYAxis = 1, cZAxis = 2 };

  CBoundingVolHierarchy( std::vector<VZone*> ZoneList, bool bIncludeInvalidZones, bool bVerbose = false, int LogLevel = 0 );
  virtual ~CBoundingVolHierarchy( void ) {
    if( m_pLeftLeaf )
      delete m_pLeftLeaf;

    if( m_pRightLeaf )
      delete m_pRightLeaf;
  }

  // copy ctor
  CBoundingVolHierarchy( const CBoundingVolHierarchy& /*rhs*/ ) = default;
  CBoundingVolHierarchy& operator=( const CBoundingVolHierarchy& /*rhs*/ ) = default;

  // construct actual tree
  virtual void BuildTree();

  // Check if the point CurPoint + t * CurDir is inside the BVH and return the corresponding zone or a nullptr
  // NOTE: if m_bIncludeInvalidRegions is true then also regions without valid bbox are checked
  // NOTE2: we have to pass the current engine in the intersection/inside functions because each thread can have its own engine
  //        which might be in a different state!!
  virtual VZone*    inside( GeometryEngine* pEngine, const Vector& CurPoint, const Vector& CurDir, double t, double eps = SMALL3D4 );

  // intersect a ray with the BVH and return nullptr or the intersected zone
  virtual VZone* intersectRay( GeometryEngine* pEngine, Ray* pCurRay );
  virtual VZone* intersect( GeometryEngine* pEngine, const Vector& Pos, const Vector& Dir, double& tmin, double tmax, VZone* pCurZone = nullptr );

  // intersect ray with regions that have invalid bounding boxes but are still stored in the tree as they are part of the original geometry
  virtual VZone* intersectRayWithInvalidZones( GeometryEngine* pEngine, Ray* pCurRay );
  virtual VZone* intersectWithInvalidZones( GeometryEngine* pEngine, const Vector& Pos, const Vector& Dir, double& tmin, double tmax, VZone* pCurZone = nullptr, CBody* pBody = nullptr );

  virtual VZone* insideInvalidZones( GeometryEngine* pEngine, const Vector& Pos, const Vector& Dir, double t, double eps );

  unsigned long     GetTotalNrOfLeaves() const { return m_NrOfLeaves; };

  unsigned long     GetTotalNrOfZones() const { return m_NrOfValidZones + m_pInvalidBBoxZones.size(); };
  unsigned long     GetNrOfInvalidZones() const { return m_pInvalidBBoxZones.size(); }
  unsigned long     GetNrOfValidZones() const { return m_NrOfValidZones; }
  unsigned long     GetNodeListSize() const { return m_pNodeList.size(); }

  virtual void      PrintTree( const std::string& FileName = "" );

  // return bounding box
  const BBox&       bbox() { return m_BoundingBox; }

protected:

	CBoundingVolHierarchy( VZone* pGeoRegion );
	CBoundingVolHierarchy( CBoundingVolHierarchy* pNode1, CBoundingVolHierarchy* pNode2 );

  // build a branch starting at ppStartNode using NrOfSubBranchNodes which will all be enclosed by one new bounding volume
	virtual CBoundingVolHierarchy* BuildBranch( VZone** ppStartNode, size_t NrOfSubBranchNodes, int Axis = cXAxis );


	// determine the number of nodes from a subset (given by ppStartNode + NrOfNodes) that have a centroid coordinate < PivotVal with respect to a specific axis.
	static  size_t    QSplit( VZone** ppStartNode, size_t NrOfNodes, double PivotVal, int Axis );

  // retrieve splitting axis and point on this axis
  virtual void GetSplittingPos( VZone* const * ppNodeList, const size_t NrOfNodes, double& Pivot, int& Axis );

  // Check if the point CurPoint + t * CurDir is inside the current region
  VZone* Contains( GeometryEngine* pEngine, const Vector& CurPoint, const Vector& CurDir, double t, double eps );

  // output current leaf
  void              PrintLeaf( int Level = 0 );
	void              Indent( int Level ) {
		for( int i = 0; i < Level * 3; ++i )
			std::cerr << " ";
	}

  // optimized intersection that avoids pre-calculating constants like the inverse direction
	VZone* intersectRay( GeometryEngine* pEngine, Ray* pCurRay,  const Vector& InvDir, double& tmin, double tmax );
  VZone* intersect( GeometryEngine* pEngine, const Vector& Pos, const Vector& Dir, const Vector& InvDir, double& tmin, double tmax );

protected:
  CBoundingVolHierarchy*        m_pLeftLeaf;
  CBoundingVolHierarchy*        m_pRightLeaf;
  VZone*                        m_pZone;

  BBox                          m_BoundingBox;
  unsigned long                 m_NrOfLeaves;
  unsigned long                 m_NrOfValidZones;

  bool                          m_bIncludeInvalidZones;       // flag indicating if regions with invalid bboxes will be brute-force checked for intersections if the BVH query fails
  bool                          m_bVerbose;
  unsigned int                  m_LogLevel;

  std::vector<VZone*>           m_pNodeList;
  std::list<VZone*>             m_pInvalidBBoxZones;   // nodes without a valid bounding box

  bool                          m_bTreeBuilt;
};

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
//                                              Surface area heuristics bounding volume hierarchies
//
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

constexpr int cNrOfBuckets = 12;
struct BucketInfo {
	int m_Count = 0;
	BBox m_Box;
};

class CSAHBoundingVolHierarchy : public CBoundingVolHierarchy
{
public:
  CSAHBoundingVolHierarchy( std::vector<VZone*> ZoneList, bool bIncludeInvalidRegions, bool bVerbose = false, int LogLevel = 0 );
	virtual ~CSAHBoundingVolHierarchy( void ) = default;

// copy ctor
  CSAHBoundingVolHierarchy( const CSAHBoundingVolHierarchy& /*rhs*/ ) = default;
  CSAHBoundingVolHierarchy& operator=( const CSAHBoundingVolHierarchy& /*rhs*/ ) = default;


protected:

	CSAHBoundingVolHierarchy( VZone* pZone );
	CSAHBoundingVolHierarchy( CBoundingVolHierarchy* pNode1, CBoundingVolHierarchy* pNode2 );

  // retrieve splitting axis and point on this axis based on SAH
	virtual void GetSplittingPos( VZone* const* ppNodeList, const size_t NrOfNodes, double& Pivot, int& Axis );

  // calculate BBox of all centroids
  BBox InitCentroidBBox( VZone* const* ppNodeList, const size_t NrOfNodes );

protected:
	BucketInfo m_Buckets[cNrOfBuckets];
  BBox       m_CentroidBBox;            // bounding box of all centroids
};

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
//                                              Linear bounding volume hierarchies - based on 3D z-order sorting
//
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

struct CMortonPrimitive {
	int           m_PrimitiveIndex;
	unsigned int  m_MortonCode;
  VZone*        m_pZone;
};

class CLBVHTreelet {
public:
  CLBVHTreelet() : m_pNodes( nullptr ) {};

  virtual ~CLBVHTreelet() {
    delete m_pNodes;
    m_ZoneList.clear();
  }

  CBoundingVolHierarchy*                  m_pNodes;
  std::vector<VZone*>                     m_ZoneList;

protected:
  // prevent copies
  CLBVHTreelet( const CLBVHTreelet& rhs );
  CLBVHTreelet& operator=( const CLBVHTreelet& rhs );
};

class CLinearBoundingVolHierarchy : public CBoundingVolHierarchy
{
public:
  CLinearBoundingVolHierarchy( std::vector<VZone*> ZoneList, bool bVerbose = false, int LogLevel = 0 );
  virtual ~CLinearBoundingVolHierarchy( void ) override;

  // copy ctor
  // We intentionally do shallow copies here because in this case each copy can resort the treelet array while the respective sub-trees
  // exist only once. This allows to have a copy of the LBVHs in threaded engines which can change the order of their own attached treelets without touching
  // the original sub-trees.
  CLinearBoundingVolHierarchy( const CLinearBoundingVolHierarchy& /*rhs*/ );
  CLinearBoundingVolHierarchy& operator=( const CLinearBoundingVolHierarchy& /*rhs*/ );

  // construct actual tree
  virtual void BuildTree() override;

  // Check if the point CurPoint + t * CurDir is inside the BVH and return the corresponding zone or a nullptr
  virtual VZone* inside( GeometryEngine* pEngine, const Vector& CurPoint, const Vector& CurDir, double t, double eps = SMALL3D4 ) override;

  // intersect a ray with the BVH and return nullptr or the intersected zone
  virtual VZone* intersectRay( GeometryEngine* pEngine, Ray* pCurRay ) override;
  virtual VZone* intersect( GeometryEngine* pEngine, const Vector& Pos, const Vector& Dir, double& tmin, double tmax, VZone* pZone = nullptr ) override;

  virtual void      PrintTree( const std::string& FileName = "" ) override;

  std::list<std::shared_ptr<CLBVHTreelet> >& getTreelets() { return m_pTreelets; };
  const std::list<std::shared_ptr<CLBVHTreelet> >& getTreelets() const { return m_pTreelets; };

protected:
  std::list<std::shared_ptr<CLBVHTreelet> > m_pTreelets;               // treelets associated with zones with a valid bbox;
  // std::vector<std::shared_ptr<CLBVHTreelet> > m_pTreelets;               // treelets associated with zones with a valid bbox;
  std::vector<VZone*>                   m_pValidZones;             // zones with a valid bbox
  BBox                                  m_CentroidBBox;            // bounding box of all centroids
};

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

// BVH for bodies

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

class CBodyBVH
{
public:
  enum  { cXAxis = 0, cYAxis = 1, cZAxis = 2 };

  CBodyBVH( std::vector<CBody*> BodyList, bool bVerbose = false, int LogLevel = 0 );
  virtual ~CBodyBVH( void ) {
    if( m_pLeftLeaf )
      delete m_pLeftLeaf;

    if( m_pRightLeaf )
      delete m_pRightLeaf;
  }

  // copy ctor
  CBodyBVH( const CBodyBVH& /*rhs*/ ) = default;
  CBodyBVH& operator=( const CBodyBVH& /*rhs*/ ) = default;

  // construct actual tree
  virtual void BuildTree();

  // Check if the point CurPoint + t * CurDir is inside the BVH and return the corresponding zone or a nullptr
  // NOTE: if m_bIncludeInvalidRegions is true then also regions without valid bbox are checked
  // NOTE2: we have to pass the current engine in the intersection/inside functions because each thread can have its own engine
  //        which might be in a different state!!
  virtual CBody*    inside( const Vector& CurPoint, const Vector& CurDir, double t, double eps = SMALL3D4 );
  virtual CBody*    insideInvalidBodies( const Vector& Pos, const Vector& Dir, double t, double eps = SMALL3D4 );

  // intersect a ray with the BVH and return nullptr or the intersected zone
  virtual CBody*    distance( Ray* pCurRay );
  virtual CBody*    distance( const Vector& Pos, const Vector& Dir, double& tmin, double& tmax );

  virtual CBody*    distanceInvalidBodies( const Vector& Pos, const Vector& Dir, double& tmin, double& tmax);
  virtual CBody*    distanceInvalidBodies( Ray* pCurRay );

  unsigned long     GetTotalNrOfLeaves() const { return m_NrOfLeaves; };

  unsigned long     GetTotalNrOfZones() const { return m_NrOfValidBodies + m_pInvalidBBoxBodies.size(); };
  unsigned long     GetNrOfInvalidZones() const { return m_pInvalidBBoxBodies.size(); }
  unsigned long     GetNrOfValidZones() const { return m_NrOfValidBodies; }
  unsigned long     GetNodeListSize() const { return m_pNodeList.size(); }

  virtual void      PrintTree( const std::string& FileName = "" );

  // return bounding box
  const BBox&       bbox() { return m_BoundingBox; }

protected:

	CBodyBVH( CBody* pBody );
	CBodyBVH( CBodyBVH* pNode1, CBodyBVH* pNode2 );

  // build a branch starting at ppStartNode using NrOfSubBranchNodes which will all be enclosed by one new bounding volume
	virtual CBodyBVH* BuildBranch( CBody** ppStartNode, size_t NrOfSubBranchNodes, int Axis = cXAxis );


	// determine the number of nodes from a subset (given by ppStartNode + NrOfNodes) that have a centroid coordinate < PivotVal with respect to a specific axis.
	static  size_t    QSplit( CBody** ppStartNode, size_t NrOfNodes, double PivotVal, int Axis );

  // retrieve splitting axis and point on this axis
  virtual void GetSplittingPos( CBody* const * ppNodeList, const size_t NrOfNodes, double& Pivot, int& Axis );

  // Check if the point CurPoint + t * CurDir is inside the current region
  CBody* Contains( const Vector& CurPoint, const Vector& CurDir, double t, double eps );

  // output current leaf
  void              PrintLeaf( int Level = 0 );
	void              Indent( int Level ) {
		for( int i = 0; i < Level * 3; ++i )
			std::cerr << " ";
	}

  // optimized intersection that avoids pre-calculating constants like the inverse direction
	CBody* distance( Ray* pCurRay,  const Vector& InvDir, double& tmin, double& tmax );
  CBody* distance( const Vector& Pos, const Vector& Dir, const Vector& InvDir, double& tmin, double& tmax );

protected:
  CBodyBVH*                     m_pLeftLeaf;
  CBodyBVH*                     m_pRightLeaf;
  CBody*                        m_pBody;

  BBox                          m_BoundingBox;
  unsigned long                 m_NrOfLeaves;
  unsigned long                 m_NrOfValidBodies;

  bool                          m_bVerbose;
  unsigned int                  m_LogLevel;

  std::vector<CBody*>           m_pNodeList;
  std::list<CBody*>             m_pInvalidBBoxBodies;   // nodes without a valid bounding box

  bool                          m_bTreeBuilt;
  bool                          m_bIncludeInvalidBodies;       // flag indicating if regions with invalid bboxes will be brute-force checked for intersections if the BVH query fails
};


#endif
// vim: ts=2
