/*
 * Copyright European Organization for Nuclear Research (CERN)
 * All rights reserved
 *
 * Author:	Vasilis.Vlachoudis@cern.ch
 * Date:	10-Oct-2018
 */

#ifndef __GEOMETRY_CONEBODY_H
#define __GEOMETRY_CONEBODY_H

#include "gbody.h"
#include "cylbody.h"

// ================================= GTRCBody ==================================
class GTRCBody : public GCylinderBase {
private:
	double	h2;
public:
	GTRCBody(const std::string& aname) : GCylinderBase(aname, Type::TRC) {}

	double	Rb()	const		{ return xlen; }
	double	Ra()	const		{ return ylen; }
	void	Rb(const double r)	{ xlen = r; }
	void	Ra(const double r)	{ ylen = r; }

	void	set(const double x,  const double y,  const double z,
		    const double hx, const double hy, const double hz,
		    const double rb, const double ra);
	void	set(const Vector& pos, const Vector& h,
		    const double rb, const double ra)
			{ set(pos.x, pos.y, pos.z, h.x, h.y, h.z, rb, ra); }
	virtual	void	set(int n, double*);
	virtual	int	get(double*) const;
	virtual	int	nodes()		const	{ return 2; }
	virtual	int	closest(const Point& r, const double d, const Vector& w) const;
	virtual void	move(int item, const Point& r, const Vector& w);

	// FIXME
	//virtual	double	distance(const double x, const double y, const double z) const;

	// Dimension labels and show flags
	// @return NULL not to show anything otherwise label
	virtual	const char*	showX()	const { return "Rb"; }
	virtual const char*	showY()	const { return "Ra"; }
	virtual const char*	showZ()	const { return "H"; }

protected:
	virtual	void	createQuads();
	virtual void	createMesh()	{ createEllConeMesh(Rb(),Ra(), Rb(),Ra()); }
	virtual	void	_updateBBox();
#ifdef USE_OBBOX
	virtual	void	_updateOBB(bool in=false);
#endif
}; // GTRCBody

// ================================= GConeBody =================================
class GConeBody : public GCylinderBase {
protected:
	int	k;		// 0: both sides, >0 positive side only, <0 negative side only
public:
	GConeBody(const std::string& aname, const Type atype)
		: GCylinderBase(aname, atype)
			{ zlen = infinite; }

	Vector	direction() const;

	double	t2()	const		{ return xlen; }
	void	t2(double at2)		{ xlen = at2; }
	virtual	void	set(int n, double*);
	virtual	int	get(double*) const;
	virtual	int	closest(const Point& r, const double d, const Vector& w) const;
	virtual void	move(int item, const Point& r, const Vector& w);
	virtual	void	rotate(const double angle, const Vector& axis);

	// FIXME
	//virtual	double	distance(const double x, const double y, const double z) const;

	// Dimension labels and show flags
	// @return NULL not to show anything otherwise label
	virtual	const char*	showX()	const {
				switch (type()) {
					case Type::XEC: return "Ry";
					case Type::YEC: return "Rz";
					case Type::ZEC: return "Rx";
					default:
						return "R";
				}
			}
	virtual const char*	showY()	const {
				switch (type()) {
					case Type::XEC: return "Rz";
					case Type::YEC: return "Rx";
					case Type::ZEC: return "Ry";
					default:
						return NULL;
				}
			}
	virtual const char*	showZ()	const { return NULL; }

protected:
	virtual	void	createQuads();
	virtual void	createMesh();
	virtual	void	_updateBBox();
#ifdef USE_OBBOX
	virtual	void	_updateOBB(bool in=false);
#endif
	virtual	Location _locationWrt(const GBody* body) const;
}; // GConeBody

#endif
