/*
 * Copyright European Organization for Nuclear Research (CERN)
 * All rights reserved
 *
 * Author:	Vasilis.Vlachoudis@cern.ch
 */

#ifndef __GEOMETRY_ENGINE_H
#define __GEOMETRY_ENGINE_H

#include <vector>

#include "ray.h"
#include "cbody.h"
#include "vzone.h"
#include "geometry.h"
#include "tool/stat.h"
#include "math/random.h"

class Ray;
class GeometryKernel;
class CBoundingVolHierarchy;

/* ============================= ZoneOfPoint ========================== */
// Scanning of region information
class ZoneOfPoint {
public:
static	const int MAXZONE = 5;
public:
	int	nMax;			// Maximum number of regions to fill
	int	n, nOld;		// New/Old zones found
	double	x,  y,  z;		// Position		// FIXME convert to Vector()
	double	dx, dy, dz;		// Direction		// FIXME convert to Vector()
	VBody*	body;			// Body to which the segments belong
	std::vector<Segment>* V;	// Vertices array of body where point belongs to
	int	from, to;		// list of vertices/bodies V[fromId:toId] that
					// are involved in the search
	VZone*	zone[MAXZONE];		// new zones (only first zone from each region)
	VZone*	zoneOld[MAXZONE];	// old zones (only first zone from each region)
	int	nRegions;		// New/Old regions found

public:
	ZoneOfPoint() : nMax(MAXZONE), n(0), nOld(0) {}
	void zero() {
			n = nOld  = 0;
			from = to = 0;
			nRegions  = 0;
		}
	void copy2Old() {
			// copy previous search
			if (n>0) memcpy(zoneOld, zone, n*sizeof(VZone*));
			nOld = n;
			n = 0;
			nRegions = 0;
		}

	void init()		{ n = nRegions = 0; }
	bool more()	const	{ return n < nMax; }
	void add(VZone* zz) {
			for (int i=0; i<n; i++)
				if (zone[i]->region() == zz->region()) return;
			nRegions++;
			zone[n++] = zz;
			assert(n<=nMax);
		}

	bool hasZone(const VZone* zo) const {
			for (int i=0; i<n; i++)
				if (zone[i] == zo) return true;
			return false;
		}
}; // ZoneOfPoint

/* ========================== GeometryEngine ========================== */
/** Geometry Engine class */
class GeometryEngine {
static	const int MAXCLIP = 5;
protected:
const	Geometry*	_geometry;	/** father geometry object	*/
	GeometryKernel*	_kernel;	/** geometry kernel object	*/
public:
	std::vector<CBody>	bodies;		/** body array			*/
	//Cache<CBody3D>	cache;		//* bodies caching

	// FIXME used by the volume calculator, it should be moved to the workers!
	Random	random;			/** random number generator	*/

protected:
	size_t	nClipBodies;		/** number of clip bodies	*/
	int	clipBodiesId[MAXCLIP];	/** clipping bodies		*/
	bool	clipNegative[MAXCLIP];	/** negative side of clip body	*/
	std::vector<int> projectBodiesId;	/** projection bodies	*/

	int	lastIRURCheckID;
	Array<CBodyOrderAccel>		irurAccel;	/** intersectRayUndefinedRegion accelerator structure */
	std::vector<CBodyOrderAccel>	zoneAccel;	/** zone accelerator structure	*/

	int	gBodyCheckId;		/** checkid for bodies		*/
	int	gBodyMaxCheckId;	/** checkid for bodies maximum previous value*/

	std::vector<VZone*>	zonesSorted;	/** sorted list of zones	*/
	std::vector<VZone*>	bvhCache;	/** cache of last used zones retrieved from BVH */

	CBoundingVolHierarchy* _boundingVolTree; /** hierarchical bounding volume accelerator*/

public:
	IFSTAT(Stat	stat;)		/** Statistics			*/

public:
	GeometryEngine(const Geometry* geo=NULL, GeometryKernel* k=NULL);
	~GeometryEngine();

	void	cleanup();

	// get/set
	void	geometry(const Geometry* geo)	{ _geometry = geo; }
	void	kernel(GeometryKernel* v)	{ _kernel = v; }
	ViewPort& view() const;

	/* bodies */
	// XXX WARNING unprotected for limits. Check before
	CBody*	getBody(const int id)		{ return &bodies[id]; }
	CBody*	getBody(const GBody* body)	{ return &bodies[body->id()]; }
	CBody*	getBody(const VBody* body)	{ return &bodies[body->id()]; }
	CBody*	getBody(const char* name) {
			const GBody* body = _geometry->getBody(name);
			return body?getBody(body):NULL;
		}

	/* zones */
const	std::vector<VZone*>& getAllSortedZones() const { return zonesSorted; }
	void	tightenAllZonesBBoxes(const Vector& minPos, const Vector& maxPos, const BBox& maxBox);

	/* regions */
	VRegion* getRegion(const int id);

	/* clip body */
	void	clipBodyClear()			{ nClipBodies = 0; }
	void	clipBodyAdd(size_t id) {
			if (id<bodies.size() && nClipBodies<MAXCLIP-1) {
				clipBodiesId[nClipBodies] = id;
				clipNegative[nClipBodies] = false;
				nClipBodies++;
			}
		}
	void	clipBodyNegative(size_t id) {
			if (id<nClipBodies)
				clipNegative[id] = true;
		}
	int	clipBody(size_t i)	const	{
			return i<nClipBodies? clipBodiesId[i] : -1;
		}
	size_t	clipBodyCount()		const	{ return nClipBodies; }

	/* project body */
	void	projectBodyClear()		{ projectBodiesId.clear(); }
	void	projectBodyAdd(size_t id) {
			if (id<bodies.size())
				projectBodiesId.push_back(id);
		}
	int	projectBody(size_t i)	const {
			return i<projectBodiesId.size()? projectBodiesId[i] : -1;
		}
	size_t	projectBodyCount()	const	{ return projectBodiesId.size(); }

	/** Derive */
	void	derive();

	/** attach BVH to this engine*/
	void	boundingVolTree( CBoundingVolHierarchy* pBVH ) { _boundingVolTree = pBVH; };
	CBoundingVolHierarchy* boundingVolTree() const { return _boundingVolTree; };

	/* 2D scan functions */
	/** @param x,y,z     starting position (I)
	  * @param u,v,w     direction vector  (I)
	  * @param zone      current zone as a hint (I)
	  * @return 'zone' found at location for the kernel universe
	  */
	VZone*	where2D(const double x, const double y, const double z,
			const double u, const double v, const double w,
			VZone* zone=NULL);
	VZone*	where2D(const Vector& p, const Vector& d, VZone* zone=NULL)
			{ return where2D(p.x,p.y,p.z, d.x,d.y,d.z, zone); };
	int	where2D(ZoneOfPoint* p);

	/* 3D ray methods */
	/** @param x,y,z     starting position (I)
	  * @param u,v,w     direction vector  (I)
	  * @param t         starting distance(I), on output hit distance (O)
	  * @param universe  universe to scan regions from (I)
	  * @param zone      current zone as a hint (I)
	  * @param body      last cross body as hint (I)
	  * @return 'zone' that crossed at distance 't'
	  */
	VZone*	whereRay(const double x, const double y, const double z,
			 const double u, const double v, const double w,
			 const double t, const int universe,
			 VZone* zone=NULL, CBody* body=NULL);

	VZone*	whereRay(const Point& p,  const Vector& d,
			 const double t,  const int universe,
			VZone* zone=NULL, CBody* body=NULL)
			{ return whereRay(p.x,p.y,p.z, d.x,d.y,d.z, t, universe, zone, body); }

	/** scan ray using a segment and update the segment zone and distance */
	VZone*	whereRay(RaySegment& segment, VZone* zone=NULL, CBody* body=NULL) {
				segment.zone = whereRay(segment.pos,
							segment.dir,
							segment.tmin,
							segment.universe,
							zone,
							body);
				return segment.zone;
			}

	bool	intersectRay(Ray* ray, bool step);
	bool	intersectRayUndefinedRegion(Ray* ray);

	/** Increase body check id. Needed when we want to calculate for the bodies
	 *  a new location
	 */
	int	incBodyCheckId()	{
			int old = gBodyCheckId;
			gBodyCheckId = ++gBodyMaxCheckId;
			//cache.inc();
			return old;
		}
	void	bodyCheckId(int value)	{ gBodyCheckId = value; }
	int	bodyCheckId()	const	{ return gBodyCheckId; }

	/* Info */
	size_t	bodiesMemory()	const;
	size_t	memory()	const;
	void	printMemory()	const;

	bool	applyClip( Ray* ray );
	bool	applyProject( Ray* ray, const double prevT );

	void	check4clip( Ray* ray );
	void	check4project( Ray* ray );

private:
	void    fillCBodyZoneReferences(VZone* zone);

	int	intersectVoxel(Ray* ray);
	bool	intersectLattice(Ray* ray);
	bool	intersectRepeatedLattice(Ray* ray);

	void	fillIRURegionAccel(Ray* ray);

friend class VZone;
friend class VRegion;
}; // GeometryEngine

#endif
