/*
 * Copyright European Organization for Nuclear Research (CERN)
 * All rights reserved
 *
 * Adapted from halotrace written by C. Theis, 2007
 *
 * Author:	Christian.Theis@cern.ch
 *
 */

#ifndef FILTER_H
#define FILTER_H

#include <cmath>
#include <vector>
#include <cassert>
#include "math/vector2d.h"

class CFilter
{
public:
  enum { cNoLUT = 0};
  CFilter( double XWidth, double YWidth, int NrOfSamples );     // if NrOfSamples = 0 --> no LUT is used but direct evaluation
  virtual ~CFilter(void) {};
  virtual double Evaluate( double x, double y) const = 0;

  // access to precomputed filter values
  virtual double EvaluateLUT( __attribute__( (unused) )double x, __attribute__( (unused) )double y) { assert( false ); return 0.0f; };

  void SetWidth( double XWidth, double YWidth );
  void GetWidth( double& X, double& Y ) const { X = m_Width.x; Y = m_Width.y;  };
  const Vector2D& GetWidth() const { return m_Width; };
  const Vector2D& GetDelta() const { return m_Delta; };

protected:
  virtual void  InitLUT() {};   // nothing to be done

protected:
  int       m_NrOf1DSamples;
  int       m_TotalSamples;
  double    m_Weight;
  Vector2D  m_Width;
  Vector2D  m_Delta;
  double    m_Spacing;
  double    m_InvWidthX;
  double    m_InvWidthY;

  std::vector<std::vector< double > > m_FilterData;       // filter LUT for precomputed filters

};

class CMitchellFilter : public CFilter
{
public:
  CMitchellFilter( double B = 0.4, double C = 0.4, double XWidth = 1.0, double YWidth = 1.0, int NrOfSamples = cNoLUT ) : CFilter( XWidth, YWidth, NrOfSamples ), m_B( B ), m_C( C ) {};
  virtual ~CMitchellFilter( void ) = default;

  double Evaluate( double x, double y ) const override { return Mitchell1D( x * m_InvWidthX ) * Mitchell1D( y * m_InvWidthX ); };

protected:
  double Mitchell1D( double x ) const {
    x = fabs( 2.0 * x );
    if( x > 1.0 )
      return ((-m_B - 6 * m_C) * x * x * x + (6 * m_B + 30 * m_C) * x * x + (-12 * m_B - 48 * m_C) * x + (8 * m_B + 24 * m_C)) * (1.0 / 6.0);
    else
      return ((12 - 9 * m_B - 6 * m_C) * x * x * x + (-18 + 12 * m_B + 6 * m_C) * x * x + (6 - 2 * m_B)) * (1.0 / 6.0);
  }

protected:

  double m_B;
  double m_C;
};

class CLanczosSincFilter : public CFilter
{
public:
  CLanczosSincFilter( double Tau = 3, double XWidth = 1.0, double YWidth = 1.0, int NrOfSamples = cNoLUT ) : CFilter( XWidth, YWidth, NrOfSamples ), m_Tau( Tau ) {};
  virtual ~CLanczosSincFilter() = default;

  double Evaluate( double x, double y ) const override { return Sinc1D( x * m_InvWidthX ) * Sinc1D( y * m_InvWidthX ); };

protected:
  double Sinc1D( double x ) const {
    x = fabs( x );
    if(x < 1e-5)
      return 1.0;
    if( x > 1.0)
      return 0.0;

    x *= M_PI;
    double Sinc = sinf( x * m_Tau ) / (x * m_Tau);
    double l = sinf( x ) / x;
    return Sinc * l;
  }

protected:
  double  m_Tau;
};

#endif
