/*
 * Copyright European Organization for Nuclear Research (CERN)
 * All rights reserved
 *
 * Author:	Vasilis.Vlachoudis@cern.ch
 * Date:	04-Feb-2010
 *
 */

#include "geometry.h"

#include <stdarg.h>

#include <sstream>

#include "gbody.h"
#include "tool/array.h"
#include "tool/token.h"
#include "tool/timer.h"
#include "math/vector.h"
#include "math/matrix4.h"
#include "graph/painter.h"
#include "geometry/regionbody.h"

using std::cout;
using std::cerr;
using std::endl;

bool Geometry::developer = false;

/* ============================== Geometry ================================= */

/** Geometry */
Geometry::Geometry() :
	editBody(NULL),
	editRegion("<editreg>")
{
	_mode = Mode::Fluka;
	gridStep     =  3;
	lighterLevel =  0;
	axisLen      = 25;
	_msg[0]      =  0;

	// nlights
	deflights    =  0;
	nlights      =  0;
	_generation  =  0;

	_backgroundColor = 0x707070;
	_transparentColor= 0x707070 | FLAG_TRANSPARENT;
	errorColor       = 0xFF0000;
	gridTextColor    = 0xA0A0A0 | FLAG_FINAL;
	labelColor       = 0xA05010 | FLAG_FINAL;
	latticeColor     = 0xFFFFFF;
	messageColor     = 0xFF9000;
	paletteTextColor = 0x000000;
	regionColor      = 0x000000;
	regionErrorColor = 0xFFB0B0;
	selectColor      = 0xFF00FF;
	titleColor       = 0xD0D000;
	vertexColor      = 0xFFFF00;
	visibleColor     = 0x7F007F;
	voxelColor       = 0xFFFFFF;
	zoneColor        = 0x606060;
	makeLatticeColor();
	makeDefaultLights();

	_errors         = nullptr;		// use default cerr

	pthread_rwlock_init(&rwlock, NULL);
	pthread_mutex_init(&mutexEdit, NULL);
} // Geometry

/** ~Geometry */
Geometry::~Geometry()
{
	// clean up local structures
	cleanup();
	voxel.cleanup();
	pthread_rwlock_destroy(&rwlock);
	pthread_mutex_destroy(&mutexEdit);

	for (auto&& iter : _bitmapPool)
		delete iter.second;
} /* ~Geometry */

/** cleanup everything. All threads must be stopped */
void Geometry::cleanup()
{
	delRegions();
	delBodies();
	delMaterials();

	lockEdit();
	editBody = NULL;
	editRegion.clear();
	unlockEdit();
	// Do not cleanup the voxel maybe we needed latter
} // cleanup

/** makeLatticeColor */
void Geometry::makeLatticeColor()
{
	latticeColor |= FLAG_LATTICE;
	voxelColor   |= FLAG_VOXEL;
} // makeLatticeColor

/** add a new body of type 'type'
 * @param name  name of body to add
 * @param type  type of body
 * @return GBody object
 */
GBody* Geometry::addBody(const std::string& name, const std::string& type)
{
	if (bodiesMap.find(name) != bodiesMap.end())
		return nullptr;
	GBody* body = GBody::newBody(name, type);
	body->id(bodies.size());
	body->generation(nextGeneration());
	bodies.push_back(body);
	bodiesMap[body->name()] = body;
	return body;
} // addBody

/** add a new #region as a GBody to be used in zone expresions
 * WARNING: in the bodiesMap the body is with a leading '#'
 *          while the GBody->name() doesn't have the leading '#'
 */
GBody* Geometry::addRegionBody(const std::string& name)
{
	std::string bodyname = '#'+name;
	auto bodyit = bodiesMap.find(bodyname);
	if (bodyit != bodiesMap.end())
		return bodyit->second;
	GBody* body = new GRegionBody(bodyname);
	body->id(bodies.size());
	body->generation(nextGeneration());
	bodies.push_back(body);
	bodiesMap[bodyname] = body;
	return body;
} // addRegionBody

/** clone an existing body named 'src' with new name 'name'
 * @param src    source body to be cloned
 * @param name   new name of cloned body
 * @return new cloned body
 */
GBody* Geometry::cloneBody(const GBody* src, const std::string& name)
{
	GBody* body = src->clone(name);
	body->id(bodies.size());
	body->generation(nextGeneration());
	bodies.push_back(body);
	if (bodiesMap.find(body->name()) != bodiesMap.end())
		errorMsg(sformat(">w> Body '%s' is duplicate",body->cname()));
	bodiesMap[body->name()] = body;
	return body;
} // cloneBody

/** delBodies */
void Geometry::delBodies()
{
	for (auto body : bodies)
		delete body;
	bodies.clear();
	bodiesMap.clear();
} // delBodies

/** invalidateBody */
void Geometry::invalidateBody(GBody* body)
{
	assert(body);
	body->updateBBox(true);		// recalculate bounding box
	body->generation(nextGeneration());
#ifdef ZONES_AS_SET
	for (auto* zone : body->zones) {
#else
	ArrayIterator<GZone*> iter(body->zones);
	while (iter) {
		GZone* zone = iter++;
#endif
		zone->generation(nextGeneration());
		zone->region->generation(nextGeneration());
		zone->updateBBox(nullptr,true); // force recalculation of bounding boxes
	}
} // invalidateBody

/** addRegion */
/** add a new region
 * @param name  name of region to add
 * @return GRegion object
 */
GRegion* Geometry::addRegion(const std::string& name)
{
	GRegion* region = new GRegion(name);
	region->id(regions.size());
	region->generation(nextGeneration());
	// XXX we could use the region->id to check if the region
	// is already defined
	regions.push_back(region);
	if (regionsMap.find(region->name()) != regionsMap.end())
		errorMsg(sformat(">w> Region '%s' is duplicate",region->cname()));
	regionsMap[region->name()] = region;
	return region;
} // addRegion

/** clone an existing region named 'src' with new name 'name'
 * it will clone as well all bodies in the original expression of the region
 * @param src  source region to be cloned
 * @return new cloned region
 */
GRegion* Geometry::cloneRegion(const GRegion* src, const std::string& name)
{
	GRegion* region = src->clone(name);
	region->id(regions.size());
	region->generation(nextGeneration());
	regions.push_back(region);
	if (regionsMap.find(region->name()) != regionsMap.end())
		errorMsg(sformat(">w> Region '%s' is duplicate",region->cname()));
	regionsMap[region->name()] = region;

	// Clone expression by cloning all bodies as well
	for (GZone* zone : src->zones()) {
		GZone* newzone = region->addZone(zone->format());
		for (int ib=0; ib<zone->size(); ib++) {
			GBody* body = (*zone)[ib];
			if (body->isBody()) {
				// construct a new body name.
				// Warning of clone of clones of bodies, use the original name
				std::string newname = sformat("%s@%s",
								body->clonedFrom()->cname(),
								region->cname());
				GBody* newbody = getBody(newname);
				if (newbody)
					*newzone << newbody;
				else
					*newzone << cloneBody(body, newname);
			} else
				*newzone << body;
		}
	}

	return region;
} // cloneRegion

/** clone all region bodies to be prepared for a transform
 * @param region    region to clone bodies
 * @return number of bodies cloned
 * */
int Geometry::cloneRegionBodies(const GRegion* region)
{
	// Clone expression by cloning all bodies as well
	int count = 0;
	for (GZone* zone : region->zones()) {
		for (int ib=0; ib<zone->size(); ib++) {
			GBody* body = (*zone)[ib];
			if (body->isBody()) {
				// construct a new body name.
				// Warning of clone of clones of bodies, use the original name
				std::string newname = sformat("%s@%s",
								body->clonedFrom()->cname(),
								region->cname());
				GBody* newbody = getBody(newname);
				if (newbody)
					(*zone)[ib] = newbody;
				else {
					count++;
					(*zone)[ib] = cloneBody(body, newname);
				}
			}
		}
	}
	return count;
} // cloneRegionBodies

/** delRegions */
void Geometry::delRegions(bool fast)
{
	for (auto region : regions) {
		region->clear(fast);
		delete region;
	}
	regions.clear();
	regionsMap.clear();
} // delRegions

/** addZone to region */
GZone* Geometry::addZone(GRegion* region, GZone::Format aformat)
{
	GZone* zone = region->addZone(aformat);
	zone->generation(nextGeneration());
	return zone;
} // addZone

/** add2exp add token to regions expression
 * @param region	region to add token to
 * @param token		token to be added either body name or + - | ()
 * @param err		returned error string in case of problem
 * @return true on success, false otherwise
 */
bool Geometry::add2exp(GRegion* region, const char* token)
{
	assert(region);
	assert(token);
	GBody* body = getBody(token);
	return region->add2exp(token, body);
} // add2exp

/** addMaterial */
Material* Geometry::addMaterial(const std::string& name)
{
	Material* material = new Material(name);
	material->id(materials.size());
	materials.push_back(material);
	if (materialsMap.find(material->name()) != materialsMap.end())
		errorMsg(sformat(">w> Material '%s' is duplicate", material->cname()));
	materialsMap[material->name()] = material;
	return material;
} // addMaterial

/** delMaterials */
void Geometry::delMaterials()
{
	for (size_t i=0; i<materials.size(); i++)
		delete materials[i];
	materials.clear();
	materialsMap.clear();
	//materialsTree.destroyTree();
} // delMaterials

/** addLight to the light list
 * @param l	light structure to add
 * @return true on error, false otherwise
 */
bool Geometry::addLight(const Light& l)
{
	if (nlights >= MAXLIGHT) return true;
	light[nlights++] = l;
	return false;
} // addLight

/** make default lights. The default lights are created at initialization
 * and they have indices from 0..deflights-1 and will remain unchanged
 * The user lights will be indexed as deflights .. lights
 */
void Geometry::makeDefaultLights()
{
	// http://www.secondpicture.com/tutorials/3d/three-point_lighting_in_3ds_max_01.html
	nlights = 0;
	light[nlights].type     = Light::Type::Sun;		// sun light
	light[nlights].relative = true;
	light[nlights].dir      = Vector(5.,-8.,-10.);
	light[nlights].distance = INFINITE;
	light[nlights].power    = 0.8;
	light[nlights].spec     = true;
	light[nlights].falloff  = Light::FallOff::Constant;
	light[nlights].shadow   = true;
	light[nlights].nsamples = 1;
	light[nlights].radius   = 0;

	nlights++;

	light[nlights].type     = Light::Type::Sun;		// fill light
	light[nlights].relative = true;
	light[nlights].dir      = Vector(-1.,1.,-1.);
	light[nlights].distance = INFINITE;
	light[nlights].power    = 0.25;
	light[nlights].spec     = false;
	light[nlights].falloff  = Light::FallOff::Constant;
	light[nlights].shadow   = false;
	light[nlights].nsamples = 1;
	light[nlights].radius   = 0;
	nlights++;


	light[nlights].type = Light::Type::Sun;		// back light
	light[nlights].relative = true;
	light[nlights].dir      = Vector(-4.,-2.,6.);
	light[nlights].distance = INFINITE;
	light[nlights].power    = 0.8;
	light[nlights].spec     = false;
	light[nlights].falloff  = Light::FallOff::Constant;
	light[nlights].shadow   = false;
	light[nlights].nsamples = 1;
	light[nlights].radius   = 0;
	nlights++;


	for (int i=0; i<nlights; i++)
		light[i].dir.normalize();

	deflights = nlights;
} // makeDefaultLights

/** parse region from an expression string
 * @param region        region to fill
 * @param token         token class to split stream into tokens
 * @param convert       (I) convert zones into STD or RPN otherwise keep in NORM
 * @param zoneSep       (O) return array with position pointers where zones are
 *                      split where zones are
 * @return true on errors
 */
bool Geometry::parseRegion(GRegion& region, Token& token,
			bool convert, std::vector<int>* zonesSep)
{
	token.setAlpha(".$@",Token::AlphaNumeric::SUFFIX);	// set additional identifier symbols
	if (mode() == Mode::Penelope) {
		// treat numbers as identifiers
		token.allow_sign = false;
		token.setAlpha("0123456789",Token::AlphaNumeric::ANY);
	}
	GZone* zone = (region.size()? region[region.size()-1] : nullptr);
	GZone::Format fmt = GZone::Format::PRODUCT;		// initial guess
	bool err   = false;
	bool hash  = false;		// #region=complement of a region
	//bool circumflex = false;	// #region=complement of outer-shell of a region
	int  depth = 0;		// parenthesis depth
	if (zone) {
		// calculate depth for a pre-existing zone
		for (int i=0; i<zone->size(); i++)
			if ((*zone)[i] == &GBody::tleft) {
				depth++;
				fmt = GZone::Format::RPN;
			} else
			if ((*zone)[i] == &GBody::tright)
				depth--;
	} else
		zone = region.addZone(GZone::Format::NORMAL);

	while (token.next()) {
		//DUMP(cout	<< "Token " << (int)token()
		//		<< " '" << token.string() << "'" << endl);
		switch (token()) {
			case Token::Type::IDENT: {
				GBody* body = nullptr;
				if (hash) {		// # region present
					body = addRegionBody(token.string());
					fmt = GZone::Format::RPN;
				} else {
					body = getBody(token.ident());
					if (body==nullptr) {
						errorMsg(sformat(
							"Region '%s:%d' invalid body '%s'",
							region.cname(), region.size(),
							token.string()));
						err = true;
					}
				}
				*zone << (body!=nullptr? body : &GBody::tnil);
				hash = false;
				}
				break;

			case Token::Type::NUMBER: {
				assert(In(mode(), Mode::Mcnp,
						Mode::Phits,
						Mode::Penelope));
				GBody* body = nullptr;
				if (hash) {		// # region present
					if (token.isInteger()) {
						int n = token.integer();
						if (n<0) {
							errorMsg(sformat(
								"Cell '%s:%d' invalid cell '#%s'",
								region.cname(), region.size(),
								token.string()));
							err = true;
						}
						if (In(mode(),	Mode::Mcnp,
								Mode::Phits,
								Mode::Penelope))
							*zone << '-';
					}
					body = addRegionBody(token.string());
					fmt = GZone::Format::RPN;
				} else {		// normal surface
					if (token.isInteger()) {
						int n = token.integer();
						// FIXME do we need to check for MCNP?
						*zone << (n>0? '-' : '+');	// for MCNP inverted!!!
						body = getBody(std::to_string(abs(n)));
					}
					if (body==nullptr) {
						errorMsg(sformat(
								"Cell '%s:%d' invalid surface '%s'",
								region.cname(), region.size(),
								token.string()));
						err = true;
					}
				}
				*zone << (body!=nullptr? body : &GBody::tnil);
				hash = false;
				}
				break;

			case Token::Type::OR:
			case Token::Type::COLON:
				if (depth==0) {		// split zones
					if (not zone->empty()) {
						if (zonesSep)
							zonesSep->push_back(token.getStream().tellg());
						// Last check for format
						// (+-) +-term => RPN
						// (+-) => PRODUCT
						if (convert) {
							if (fmt == GZone::Format::PRODUCT and
							    (*zone)[0] == &GBody::tleft   and
							    (*zone)[zone->size()-1] != &GBody::tright)
								fmt = GZone::Format::RPN;
							try {
								zone->convertTo(fmt);
							} catch (std::string& ex) {
								zone->clear();
								errorMsg(sformat(
									"Region '%s:%d' %s",
									region.cname(), region.size(),
									ex.c_str()));
								err = true;
							}
						}
						zone = region.addZone(GZone::Format::NORMAL);
						fmt  = GZone::Format::PRODUCT;
					}
				} else {
					*zone << '|';
					fmt = GZone::Format::RPN;
				}
				break;

			case Token::Type::LEFT_PAREN:
				if (In(mode(), Mode::Mcnp, Mode::Phits)) {
					if (hash)
						*zone << '-';
					else
					if (zone->size())
						*zone << '+';
				}
				*zone << '(';
				hash = false;
				// set type to RPN if '(' is not the first token
				fmt  = GZone::Format::RPN;
				depth++;
				break;

			case Token::Type::RIGHT_PAREN:
				*zone << ')';
				if (--depth<0) {
					errorMsg(sformat
						("Region '%s:%d' too many closing ) parenthesis",
							region.cname(), region.size()));
					err = true;
					depth = 0;
				}
				break;

			case Token::Type::PLUS:
				if (In(mode(), Mode::Fluka, Mode::Moira))
					*zone << '+';
				else
				if (mode() == Mode::Penelope)
					*zone << '-';
				break;

			case Token::Type::MINUS:
				if (In(mode(), Mode::Fluka, Mode::Moira))
					*zone << '-';
				if (mode() == Mode::Penelope)
					*zone << '+';
				break;

			case Token::Type::HASH:
				if (hash) {
					errorMsg(sformat(
						"Cell '%s:%d' Double # operator found",
						region.cname(), region.size()));
					err = true;
				}
				if (In(mode(), Mode::Fluka, Mode::Moira, Mode::Penelope))
					// Add only for FLUKA, since for MCNP/Phits
					// it depends on what follows
					*zone << '-';
				hash = true;
				break;

			//case Token::Type::NOT:	// ^
			//	if (In(self->geometry->mode(),
			//	    Geometry::Mode::Fluka, Geometry::Mode::Moira))
			//		// Add only for FLUKA, since for MCNP/Phits
			//		// it depends on what follows
			//		*zone << &GBody::tminus;
			//	if (circumflex) {
			//		errorMsg(sformat(
			//			"Cell '%s:%d' Double ^ operator found",
			//			region.cname(), region.size()));
			//		err = true;
			//	}
			//	circumflex = true;
			//	break;

			case Token::Type::NEWLINE:	// skip
				break;

			default: {
				errorMsg(sformat("Region '%s:%d' invalid token '%s'",
					region.cname(), region.size(),
					token.string()));
				err = true;
			}
		}
	}
	DUMP(std::cout << "<parseRegion: " << region << std::endl);

	// FIXME creates an error when we are parsing incomplete regions
	// in multiple lines/passages like in nox
	//if (depth!=0) {
	//	errorMsg(
	//		sformat("Region '%s:%d' too many opening '(' parenthesis",
	//			region.cname(), region.size()));
	//} else
	//	if (region.type() != GRegion::Type::VOXEL)
	//		errorMsg(sformat("Region '%s' empty expression",
	//				region.cname()));
	// Last check for format
	// (+-) +-term => RPN
	// (+-) => PRODUCT
	if (zone->empty())
		region.delZone(region.size()-1);
	else
	if (convert) {
		if (fmt == GZone::Format::PRODUCT and
		    (*zone)[0] == &GBody::tleft   and
		    (*zone)[zone->size()-1] != &GBody::tright)
			fmt = GZone::Format::RPN;
		try {
				zone->convertTo(fmt);
		} catch (std::string& ex) {
			zone->clear();
			errorMsg(sformat("Region '%s:%d' %s",
					region.cname(), region.size(),
					ex.c_str()));
		}
	}
	DUMP(std::cout << ">parseRegion: " << region << std::endl);
	return err;
} // parseRegion

/** parseRegion */
bool Geometry::parseRegion(GRegion& region, const std::string& expstr, bool convert)
{
	std::stringstream stream(expstr);
	Token token(stream);
	return parseRegion(region, token, convert);
} // parseRegion

/** derive: setup all bodies and regions involved in the geometry */
int Geometry::derive()
{
	int err = 0;
	lockWrite();

	// update bounding boxes
	for (auto body : bodies)
		body->updateBBox();

	// check for touching bodies
	int touch_count = 0;
	for (unsigned i=0; i<bodies.size(); i++) {
		for (unsigned j=i+1; j<bodies.size(); j++)
			if (bodies[i]->isTouching(*bodies[j])) {
				// maintain a cycle list of touching bodies
				bodies[i]->addTouching(bodies[j]);
				errorMsg(sformat(">w> Body '%s' [%s] and '%s' [%s] are touching",
				                 bodies[i]->cname(), bodies[i]->typeStr(),
				                 bodies[j]->cname(), bodies[j]->typeStr()));
				touch_count++;
			}
		if (touch_count > 100) {
			errorMsg(">w> Too many body touching messages...");
			break;
		}
	}

	// expand regions (complement operator)
	// and setup zones of regions (bounding boxes, skip array for RPN expr)
	// Loop in reverse, for Penelope to avoid expanding the expanded regions
	for  (size_t ir=regions.size(); ir--; ) {
		auto region = regions[ir];
		if (!region->expandRegions(this)) {
			errorMsg(sformat(">e> Region '%s' cannot expand",
			                 region->cname()));
			err++;
		}
		region->setup();		// setup region (bbox, calcLength,...
		DUMP(std::cout << "Geometry::derive: " << *region << std::endl);
	}
	unlockWrite();
	return err;
} // derive

/** printInfo */
void Geometry::printInfo() const
{
	cout << "Geometry Info" << endl;
	cout << "Number of bodies: " << bodies.size() << endl;
	cout << "Number of regions: " << regions.size() << endl;
	int nzones = 0;
	for (auto region : regions)
		nzones += region->size();
	cout << "Number of zones: " << nzones << endl;
} // printInfo

/** bodiesMemory */
size_t Geometry::bodiesMemory() const
{
	size_t bmem = bodies.capacity()*sizeof(GBody*) + sizeof(bodiesMap);
	for (auto b : bodies)
		bmem += b->memory();
	return bmem;
} // bodiesMemory

/** regionsMemory */
size_t Geometry::regionsMemory() const
{
	size_t rmem = regions.capacity()*sizeof(GRegion*) + sizeof(regionsMap);
	for (auto r : regions)
		rmem += r->memory();
	return rmem;
} // regionsMemory

/** errorMsg */
void Geometry::errorMsg(std::string msg)
{
	if (_errors)
		_errors->push_back(msg);
	else
		cerr << msg << endl;
} // errorMsg

/** memory */
size_t Geometry::memory() const
{
	return   sizeof(Geometry)
	       + bodiesMemory()
	       + regionsMemory();
//	       + voxel.memory();
} // memory

/** printMemory */
void Geometry::printMemory() const
{
	cout << endl << "Geometry:" << endl;
	cout << "Bodies:\t\t" << bodies.size()   <<  endl;
	cout << "Regions:\t"  << regions.size()  <<  endl;
	cout << "Memory:"     << endl;
	cout << "\tSelf:\t"   << sizeof(Geometry) << endl;
	cout << "\tBodies:\t" << bodiesMemory()   << endl;
	cout << "\tRegion:\t" << regionsMemory()  << endl;
	cout << "\tTotal:\t"  << memory() << endl;
} // printMemory

/** load bitmap into pool */
CDIB* Geometry::loadBitmap( const std::string& fileName )
{
	CDIB* pBitmap = bitmapLoaded( fileName );
	if ( pBitmap ) return pBitmap;

	pBitmap = new CDIB;
	if( !pBitmap->OpenDIB( fileName ) ) return nullptr;

	_bitmapPool.insert( make_pair( fileName, pBitmap) );
	return pBitmap;
} // loadBitmap


/** unload bitmap from pool */
bool Geometry::unloadBitmap( CDIB* pBitmap )
{
	if ( !pBitmap || !bitmapLoaded( pBitmap) ) return false;

	auto iter = _bitmapPool.find( pBitmap->GetName() );
	if ( iter == _bitmapPool.end() ) return false;

	pBitmap->DestroyDIB();
	delete pBitmap;
	_bitmapPool.erase( iter );

	pBitmap = nullptr;
	return true;
} // unloadBitmap

/** unload bitmap from pool */
bool Geometry::unloadBitmap( const std::string& fileName )
{
	CDIB* pBitmap = bitmapLoaded( fileName );
	if ( !pBitmap ) return false;
	return unloadBitmap( pBitmap );
} // unloadBitmap


/** check if bitmap has been loaded already */
CDIB* Geometry::bitmapLoaded( const std::string& fileName )
{
	auto iter = _bitmapPool.find( fileName );
	if ( iter != _bitmapPool.end() ) return iter->second;
	return nullptr;
} // bitmapLoaded

/** check if bitmap has been loaded already */
CDIB* Geometry::bitmapLoaded( CDIB* pBitmap )
{
	for ( auto&& iter : _bitmapPool ) {
		if ( pBitmap == iter.second )
			return pBitmap;
	}

	return nullptr;
} // bitmapLoaded
