/*
 * Copyright European Organization for Nuclear Research (CERN)
 * All rights reserved
 *
 * Author: Chris.Theis@cern.ch
 * Date:   28-Jan-2021
 */

#include "qmcdebug3d.h"

#include "gregion.h"
#include "geometry.h"
#include "math/bit.h"
#include "math/bbox.h"
#include "math/randsampler.h"

//using namespace flair;
using std::cout;
using std::endl;

// =============================== Debug3DWorker ================================
#ifdef THREAD
/** operator() */
void Debug3DWorker::operator()()
{
	std::vector<GRegion*> containments;
	double randNr[2];

	Debug3DFeeder* f      = static_cast<Debug3DFeeder*>(feeder());
	const Debug3DJob* job = f->job;
	const Point& low      = job->bbox.low();
	const Vector& size    = job->bbox.size();


	while( idx < to ) {
		randNr[0] = CBaseSampler::VanDerCorputDbl( idx, 532579235 );
		randNr[1] = CBaseSampler::Sobol2Dbl( idx, 6743690439 );

		// sample coordinates using low-discrepancy series
		// for the time being we keep 1 dimension as a grid. Works fine but could be
		// improved by using Niederreiter [0,2] series in 3 dimensions or a simple LHC or jittering
		double x = low.x + size.x * randNr[0];
		double y = low.y + size.y * randNr[1];
		double z = low.z + size.z / (double)f->samples * idx;

		// do containment tests for all regions
		for( auto Iter = pRegions->begin(); Iter != pRegions->end(); ++Iter ) {
			if( (*Iter)->inside( x, y, z, 0., 0., 1.0 ) )
				containments.push_back( *Iter );
		}

		// check for undefined points
		if( containments.empty() )
			undefinedPoints.push_back( Vector( x, y, z ) );

		// handle multiple containments
		if( containments.size() > 1 ) {
			ambiguousPoints.insert( std::make_pair<Vector, std::vector<GRegion*> >( Vector( x, y, z ), std::move( containments ) ) );
		}
		std::vector<GRegion*>().swap( containments );

		idx++;
	}

} // operator

// =============================== Debug3DFeeder ================================
/** allocate workers if the requested number of threads changed */
void Debug3DFeeder::allocate()
{
	if (nworkers != pool.nthreads()) {
		if (workers) delete [] workers;
		nworkers = pool.nthreads();
		workers = new Debug3DWorker[nworkers];
		for (int i=0; i<nworkers; i++)
			workers[i].feeder(this);
	}
} // allocate

/** reset feeder */
void Debug3DFeeder::reset(const Debug3DJob* j, int nsamples, int nstep, Random& random)
{
	pool.reset();
	allocate();

	job     = j;
	idx     = 0;
	step    = nstep;
	samples = nsamples;

	// initialize workers
	for (int i=0; i<nworkers; i++)
		workers[i].reset(random.integer(), &(job->geometry->regions) );
} // reset

/** feed */
ThreadPoolWorker* Debug3DFeeder::feed(int threadId)
{
	// Check for ending condition
	if (idx>=samples) return nullptr;

	// update worker
	workers[threadId].idx = idx;
	workers[threadId].to  = idx = Min(idx + step, samples);	// next value

	return (ThreadPoolWorker*)&workers[threadId];
} // feed
#endif

// ================================== Debug3DJob ==================================
/** Stochastic debugging in 3D via containment tests
 * @param samples (I/O) I:maximum number of samples to run, O:current status
 * @return upon finishing the job contains two containers with infos related to
 *				 undefined and ambiguously defined points:
 *						std::multimap<Vector, std::vector<GRegion*> > ambiguousPoints;
 *						std::vector<Vector> undefinedPoints;
 *
 *				 For the ambiguously defined points a list of regions which include them
 *				 is provided as well
*/

/** initialize job
 * @return true if everything is ok
 */
bool Debug3DJob::init(const Geometry* geo, const BBox& regionOfInterest, int samples, int threads)
{
	geometry = geo;
	bbox = regionOfInterest;

	run_samples = samples;
	if (!isPowerOf2(run_samples))
		run_samples = roundUpPow2(run_samples);
	nsamples = 0;

	initThreads(threads);

	return true;
} // init

/** run debug job, either using the threadpool or in a single thread */
void Debug3DJob::run()
{
	startRun();

	Random random;
	nsamples = 0;

	ambiguousPoints.clear();
	undefinedPoints.clear();

#ifdef THREAD
	if (nthreads() > 1) {
		Debug3DFeeder feeder(threadpool);
		for (int n=32; n<=run_samples; n<<=1) {
			feeder.reset(this, n, 1024, random);
			if (threadpool.execute(&feeder)) break;
			if (threadpool.isStopped()) break;
			ambiguousPoints = feeder.ambiguousPoints();
			undefinedPoints = feeder.undefinedPoints();
			nsamples = n;

			DUMP(cout << nsamples
				<< " / " << run_samples << " / " << run_samples << " coverage = " << bbox.volume() / nsamples << " cm3/point"
				<< " nr. of undefined points =" << undefinedPoints.size()
				<< " nr. of ambiguous points =" << ambiguousPoints.size()
				<< endl);
		}
	} else
#endif
	{
		double randNr[2];
		std::vector<GRegion*> containments;

		int bp = 1 << 5;	// breakpoint to show progress
		const Point& low = bbox.low();
		const Vector& size = bbox.size();

		for( int n = 0; n < run_samples; n++ ) {

			randNr[0] = CBaseSampler::VanDerCorputDbl( n, 532579235 );
			randNr[1] = CBaseSampler::Sobol2Dbl( n, 6743690439 );

			// sample coordinates using low-discrepancy series
			double x = low.x + size.x * randNr[0];
			double y = low.y + size.y * randNr[1];
			double z = low.z + size.z / (double)run_samples * n;

			// do containment tests for all regions
			for( size_t i = 0; i < geometry->regions.size(); ++i ) {
				if( geometry->regions[i]->inside( x, y, z, 0., 0., 1.0 ) )
					containments.push_back( geometry->regions[i] );
			}

			// check for undefined points
			if( containments.empty() )
				undefinedPoints.push_back( Vector( x, y, z ) );

			// handle multiple containments
			if( containments.size() > 1 ) {
				ambiguousPoints.insert( std::make_pair<Vector, std::vector<GRegion*> >( Vector( x, y, z ), std::move( containments ) ) );
			}
			std::vector<GRegion*>().swap( containments );

			// Calculate temporarily for reporting
			if( ++nsamples > bp ) {

				DUMP( cout << nsamples
							<< " / " << run_samples << " coverage = " << bbox.volume() / nsamples << " cm3/point"
							<< " nr. of undefined points =" << undefinedPoints.size()
							<< " nr. of ambiguous points =" << ambiguousPoints.size()
							<< endl );
				bp <<= 1;
			}
		}

	}


	endRun();
} // run

/** spawn */
void Debug3DJob::spawn()
{
	pthread_create(&thread, NULL,
		(void*(*)(void*))Debug3DJob::runInThread, this);
} // spawn

/** runInThread */
void* Debug3DJob::runInThread( Debug3DJob* job)
{
	job->run();
	job->thread = 0;
	return NULL;
} // runInThread

// vim: ts=2
