/*
 * Copyright European Organization for Nuclear Research (CERN)
 * All rights reserved
 *
 * Author: Chris.Theis@cern.ch
 * Date:   28-Jan-2021
 */

#ifndef __GEOMETRY_QMCDEBUG3D_H
#define __GEOMETRY_QMCDEBUG3D_H

#include "gjob.h"
#include "gbody.h"
#include "math/random.h"

class BBox;
class Random;
class GRegion;
class Geometry;

//namespace flair {

// =============================== Debug3DWorker ================================
// Use QMC series to determine if a set of points is well defined within a geometry.
// Ambiguous definitions as well as holes should be detected.

#ifdef THREAD
class Debug3DJob;
class Debug3DWorker : public ThreadPoolWorker {
public:
	int	idx;			/** index of event		*/
	int	to;			/** up to which event to calc	*/
	Random	random;			/** random number generator	*/
	long	randSeed;		/** seed for QMC sequence generation **/
	const std::vector<GRegion*>*	pRegions;

	std::vector<Vector>   undefinedPoints;
	std::multimap<Vector, std::vector<GRegion*> >  ambiguousPoints;

public:
	Debug3DWorker(): idx(0), to(0) {}
virtual	void	operator ()();
	void	reset(long s, const std::vector<GRegion*>* r) {
				random.seed(s);
				randSeed = s;
				pRegions = r;
				idx = to = 0;
				undefinedPoints.clear();
				ambiguousPoints.clear();
			}
}; // class VolumeWorker

// ================================ Debug3DFeeder ================================
class Debug3DFeeder : public ThreadPoolFeeder {
public:
const	Debug3DJob*	job;		/** volume job pointer		*/
	int		samples;	/** maximum number of samples	*/

private:
	int		nworkers;	/** number of workers allocated	*/
	Debug3DWorker*	workers;	/** array of body workers	*/
	int		idx;		/** next start index		*/
	int		step;		/** step per worker		*/

public:
	Debug3DFeeder(ThreadPool& p) : ThreadPoolFeeder(p),
		  nworkers(0),
		  workers(NULL)
		  {}
	virtual	~Debug3DFeeder()		{ if (workers) delete [] workers; }

	// Initialize workers
	void	allocate();
virtual	void	reset(const Debug3DJob* j, int nsamples, int nstep, Random& random);

virtual	ThreadPoolWorker* feed(int threadId);

	/** @accumulate and return all undefined points */
	std::vector<Vector>	undefinedPoints() {
		std::vector<Vector> result;
		for( int i = 0; i < nworkers; i++ )
			result.insert( std::end(result), std::begin(workers[i].undefinedPoints), std::end(workers[i].undefinedPoints) );

		return result;
	}

	/** @accumulate and return all ambiguously defined points */
	std::multimap<Vector, std::vector<GRegion*> >	ambiguousPoints() {
		std::multimap<Vector, std::vector<GRegion*> >  result;
		for( int i = 0; i < nworkers; i++ )
			result.insert( workers[i].ambiguousPoints.begin(), workers[i].ambiguousPoints.end() );

		return result;
	}
}; // class Debug3DFeeder
#endif

// ================================== Debug3DJob ==================================
class Debug3DJob : public GeometryJob {
public:
	BBox	bbox;		/** bounding box of ROI	*/
	int		run_samples;	/** number of samples to run	*/
	int		nsamples;	/** number of samples run	*/

	std::multimap<Vector, std::vector<GRegion*> >	ambiguousPoints; /** ambiguously defined points */
	std::vector<Vector>	undefinedPoints;	/** undefined points */

public:
	Debug3DJob() : GeometryJob() {}
	virtual	~Debug3DJob() {}

	bool	init( const Geometry* geo, const BBox& regionOfInterest, int samples, int threads = 0 );

	// FIXME setting to virtual causes a coredump
	void	run();

	/** spawn a job in a background thread */
	void	spawn();
private:
static	void*	runInThread(Debug3DJob* job);
}; // class Debug3DJob

//} // namespace flair

#endif
// vim: ts=2
