/*
 * Copyright European Organization for Nuclear Research (CERN)
 * All rights reserved
 *
 * Author:	Vasilis.Vlachoudis@cern.ch
 */

#ifndef __GEOMETRY_RAY_H
#define __GEOMETRY_RAY_H

#include "geo.h"
#include "cbody.h"
#include "vzone.h"

#include "math/point.h"
#include "math/vector.h"
#include "geometry/cbody.h"


class Matrix4;

/* ============================= RaySegment =========================== */
class RaySegment {
public:
	Point	pos;			// current local system position
	Vector	dir;			// current local system direction
	VZone*	zone;			// Current location
	CBody*	body;			// entering body
const	GRegion	*region;		// lattice region for transformation matrix accessing
	double	tmin;			// range of segment
	double	tmax;			// [tmin..tmax)
	double	acc;			// range accuracy
	int	bodyCheckId;		// remember last checkId
	int	universe;		// universe the segment is in
	bool	incremental;		// true(default) if new segment starts from tmin~=0
					// when removed the distance traveled has to be
					// transferred to the parent segment

public:
	// FIXME universe should be defined even if region is NULL
	RaySegment(const GRegion* reg=NULL, int u=0) :
			pos(0.,0.,0.),
			dir(0.,0.,0.),
			zone(NULL),
			body(NULL),
			region(reg),
			tmin(SMALL3D),
			tmax(INFINITE),
			acc(SMALL3D),
			bodyCheckId(0),
			universe(u),
			incremental(true)
			{}
	RaySegment(const double x,  const double y,  const double z,
		   const double dx, const double dy, const double dz,
		   VZone* zz, int checkId=0) :
			pos(x,y,z),
			dir(dx,dy,dz),
			zone(zz),
			body(NULL),
			region(NULL),
			tmin(SMALL3D),
			tmax(INFINITE),
			acc(SMALL3D),
			bodyCheckId(checkId),
			incremental(true)
			{ universe = (zz? zz->universe() : 0); }
	RaySegment(const Point& p, const Vector& d, VZone* zz, int checkId=0) :
			pos(p),
			dir(d),
			zone(zz),
			body(NULL),
			region(NULL),
			tmin(SMALL3D),
			tmax(INFINITE),
			acc(SMALL3D),
			bodyCheckId(checkId),
			incremental(true)
			{ universe = (zz? zz->universe() : 0); }

	bool	ended()	const	{ return tmin*(1.0+acc) >= tmax; }

	// Relative hit position inside the segment-frame
	// LATTICE: are screen coordinates transformed by lattice rot-translation
	// VOXEL: are absolute coordinates
	Point	hit()	const	{ return pos + tmin*dir; }
	Point	hit(const double boost) const
				{ return pos + tmin*(1.0+boost)*dir; }

	bool	hasMatrix()	const { return region!=nullptr and region->hasMatrix(); }
const	Matrix4& matrix()	const { assert(region); return region->matrix();    }
const	Matrix4& invMatrix()	const { assert(region); return region->invMatrix(); }
}; // RaySegment

/* ================================= Ray ============================== */
/** Intersect Ray structure
 * This structure contains a list of RaySegments,
 *   the last segment of the list, the one pointed by n (segments[n]) is
 *   considered the working one and the rest in the list are the previous
 *   segments.
 */
class Ray {
public:
	enum class SkipBlack : int {
		StopOnBlack  = 0,
		Skip1stBlack = 1,
		SkipAllBlack = 2,
	};

	enum class RayType : int {
		Standard     = 0,
		Reflection   = 1,
		Transmission = 2,
	};

public:
	bool	error;			// If ray passed through an undefined region
	bool	shadow;			// calculate shadows for ray
	int	nlights;		// number of lights to use
	bool	skip_current;		// Skip current zone as if it was transparent
	bool	skip_transparent;	// Skip transparent zones
	SkipBlack skip_black;		// Skip the starting blackhole if any
	int	voxelreg;		// voxel region if any

	Vector	normal;			// keep normal at hit position

	/* Clipping body */
	bool	use_clip;		// flag if we have to use clipping
	bool	clip_insight;		// possible clip along the ray
	bool	clip_hit;		// clipping body was hit

	/* Project body */
	bool	use_project;		// flag if we have to use projection
	bool	project_insight;	// possible projection along the ray
	bool	project_hit;		// project was hit
	int	project_alpha;		// projection transparency

	int	depth;			// segment depth
	int	max_depth;		// maximum segment depth

	RayType	type;

const	static int MaxNestingLevel=20;	// maximum ray nesting level

private:
	VZone*	_prevzone;		// current upstream zone for n=0 (just before the hit)
	double	tsum;			// tmin running length of previous levels
public:
	int	n;			// lattice depth level
private:
	RaySegment segments[MaxNestingLevel];

public:
	Ray()	: shadow(false),
		  nlights(0),
		  skip_black(SkipBlack::StopOnBlack),
		  use_clip(false),
		  use_project(false),
		  depth(0),
		  max_depth(0), type( RayType::Standard )
			{init();}

	Ray(const Ray& ray) = default;
	Ray& operator = (const Ray&) = default;

	// clone
	// NOTE: Normally I'd implement this in terms of assignment op.
	//       However, I saw that the copy ctor is not really creating a deep copy
	//       but increases the ray's depth and does a fresh init().
	//       To avoid side effects I keep the compiler's default assignment
	//       op and rather create a specific clone function.
	//       CT 16/8/2019
	//
	void	clone(Ray& destRay) const {
			if (&destRay != this) {
				destRay.shadow          = shadow;
				destRay.nlights         = nlights;
				destRay.use_clip        = use_clip;
				destRay.use_project     = use_project;
				destRay.skip_black      = skip_black;
				destRay.depth           = depth;
				destRay.max_depth       = max_depth;

				destRay.voxelreg        = voxelreg;
				destRay.error           = error;
				destRay.skip_current    = skip_current;
				destRay.skip_transparent= skip_transparent;

				destRay.clip_insight    = clip_insight;
				destRay.clip_hit        = clip_hit;

				destRay.project_insight = project_insight;
				destRay.project_hit     = project_hit;
				destRay.project_alpha   = project_alpha;

				destRay._prevzone       = _prevzone;
				destRay.tsum            = tsum;
				destRay.n               = n;

				destRay.normal          = normal;
				destRay.type            = type;

				std::copy(std::begin(segments),
					  std::end(segments),
					  std::begin(destRay.segments));
			}
		}

	void	init() {
			voxelreg         = -1;
			error            = false;
			skip_current     = false;
			skip_transparent = true;
			clip_insight     = false;
			clip_hit         = false;
			project_insight  = false;
			project_hit      = false;
			project_alpha    = 0;
			_prevzone        = NULL;
			tsum             = 0.0;
			n                = -1;
			type             = RayType::Standard;
		}

	// Return true if ray segments buffer is full
	bool	isFull()	const	{ return n+1 >= MaxNestingLevel; }
	bool	isEmpty()	const	{ return n == -1; }
	bool	start()		const	{ assert(n>-1); return T() <= SMALL3D3; }
	void	incDepth()		{ depth++; }

	// prefix operator to increase the depth counter
	Ray& operator++()		{ ++depth;  return *this; };
	// postfix op implemented in terms of the prefix op
	Ray  operator++(int)		{ Ray temp = *this; ++*this; return temp; };

	// Consolidate the current segment and make _segment the new one
	bool	push(const RaySegment& _segment) {
			if (isFull()) {
				error = true;
				return true;
			}
			if (n >= 0) {
				if (_segment.incremental) {
					//assert(_segment.tmin < SMALL3D4);
					tsum += segments[n].tmin;
				}
				if (segments[0].universe == _segment.universe)
					_prevzone = segments[n].zone;
			}
			segments[++n] = _segment;
			return false;
		}

	// Discards the current segment and makes as current the previous one
	bool	pop() {
			if (isEmpty()) return true;
			if (n > 0 and segments[n].incremental) {
				tsum -= segments[--n].tmin;
			} else
				n--;
			return false;
		}

	// Return reference to any segment
	RaySegment&	segment(int idx)	{ assert(idx<=n); return segments[idx]; }
const	RaySegment&	segment(int idx) const	{ assert(idx<=n); return segments[idx]; }
	// Current segment
	RaySegment&	segment()		{ assert(n>=0); return segments[n]; }
const	RaySegment&	segment()	const	{ assert(n>=0); return segments[n]; }
	int	universe()		const	{ assert(n>=0); return segments[n].universe; }

const	Point&	pos()			const	{ assert(n>=0); return segments[0].pos; }
const	Vector&	dir()			const	{ assert(n>=0); return segments[0].dir; }

	/** @return absolute screen coordinates hit position */
	Point	hit() const {
			assert(n>=0);
			return pos() + (segments[n].tmin+tsum)*dir();
		}

	/** @return hit position including a tiny boost along the motion */
	Point	hit(const double boost) const {
			assert(boost!=0.0);
			if (boost>0.0)	// Forward boost
				return pos() + ((segments[n].tmin+tsum)*(1.0+boost))*dir();
			else		// Backward boost
				return pos() + ((segments[n].tmin+tsum)/(1.0-boost))*dir();
		}

	/** @return hit position in local object coordinates */
	Point	hitInObjCoords() const {
			Matrix4 objMatrix;
			objMatrix.make( hitBody()->body()->vectorX(),
					hitBody()->body()->vectorY(),
					hitBody()->body()->vectorZ() );
			objMatrix.setT( hitBody()->body()->position() );
			return objMatrix.inv() * hit();		// hit point in object coordinates
		}

	// set/get
	void	prevZone(VZone *z)    { assert(n==0); _prevzone = z; }
	VZone*	prevZone()	const { return _prevzone; }
	CBody*	hitBody()	const { for (int i=n; i>=0 ; i--)
						if (segments[i].body)
							return segments[i].body;
					return nullptr;
				}
	VZone*	hitZone()	const { return segment().zone; }
	VRegion* hitRegion()	const { return segment().zone->region(); }

	const Material*	hitMaterial() const {
			if ( !hitRegion() )
				return nullptr;
			// transmission rays (=exiting rays) must return the
			// material of the zone from which they exit
			// (potentially into air/vacuum = 100% transparent) in
			// order to invoke the correct calculation of
			// refraction
			if ( type != Ray::RayType::Transmission )
				return hitRegion()->region()->material();

			assert(prevZone());
			return prevZone()->region()->region()->material();
		}

	void	moveby(double a)      { segment().tmin = T()*(1.0+a) - tsum; }

	double	T()		const { return segments[n].tmin + tsum; }
	double	Tsum()		const { return tsum; }
	bool	ended()		const { return T()>=segments[0].tmax; }
	// artificially end a ray
	void	stop() {
			depth = max_depth + 1;
			segments[0].tmax = -INFINITE;
		}
}; // Ray

#endif
