/*
 * Copyright European Organization for Nuclear Research (CERN)
 * All rights reserved
 *
 * Author:	Vasilis.Vlachoudis@cern.ch
 * Date:	16-Jul-2010
 */

#include "color.h"

#include "math/bit.h"

// if wchar_t is 4 bytes use the wmemset for faster background setting
#if WCHAR_MAX > 0xffffu
//	wchar_t is 4 bytes
#	define WCHAR4
#endif

// --- Declare Basic Colors ---
const Color Color::Black(0.0,0.0,0.0);
const Color Color::Red(1.0,0.0,0.0);
const Color Color::Green(0.0,1.0,0.0);
const Color Color::Blue(0.0,0.0,1.0);
const Color Color::Magenta(1.0,0.0,1.0);
const Color Color::Turquoise(0.0,1.0,1.0);
const Color Color::Yellow(1.0,1.0,0.0);
const Color Color::White(1.0,1.0,1.0);

/* remap colors in the range from lw .. hg */
void Color32::remap(const Color32& lw, const Color32& hg)
{
	red()   = (byte)(((iR() * (hg.iR() - lw.iR()))>>8) + lw.iR());
	green() = (byte)(((iG() * (hg.iG() - lw.iG()))>>8) + lw.iG());
	blue()  = (byte)(((iB() * (hg.iB() - lw.iB()))>>8) + lw.iB());
} // remap

/* alpha color blending */
void Color32::blend(const Color32& color)
{
	byte alphaBnegA  = byteMul(color.alpha(), alpha()^0xff);	// 255-a = a^0xff
	red()   = byteMul(red(),   alpha()) + byteMul(color.red(),   alphaBnegA);
	green() = byteMul(green(), alpha()) + byteMul(color.green(), alphaBnegA);
	blue()  = byteMul(blue(),  alpha()) + byteMul(color.blue(),  alphaBnegA);
	alpha(0);
} // blend

/* ========================= HSV ================================== *

                          ^ Value
                   _______|______
                   \      |      /   Side face of hexcone HSV
                     \    |    /
                       \  |  /
                         \|/
                Green          Yellow            /
                    -------------               /
                  / \           / \            /
                /     \       /     \         / ^  Hue
              /         \   /         \          \
      Cyan  /_____________x_____________\   Red  --------->
            \       White  \            /                Saturation
              \         /    \        /
                \     /        \    /
                  \ /            \/          Top face of hexcone HSV
              Blue  ~~~~~~~~~~~~~~ Magenta   color model

 * ================================================================ */

/** getHSV - convert an RGB color to HSV
 *  Output:  h in [0,360],  s and v in [0,1], except if s = 0
 *           then h is undefined which is defined as a constant -1
 *           outside the interval [0,360]
 * @param	col	Color32 to convert to HSV
 * @param	h	return hue [0..360]
 * @param	s	return saturation [0..1]
 * @param	v	return value [0..1]
 */
void Color32::getHSV(double* h, double* s, double* v) const
{
	int maximum = Max(red(), green(), blue());
	int minimum = Min(red(), green(), blue());

	/* value */
	*v = (double)maximum/255.0;

	if (maximum != 0)
		/* saturation */
		*s = (double)(maximum - minimum) / (double)maximum;
	else
		*s = 0.0;

	if (*s == 0.0)
		*h = -1.0;		/* hue, undefined */
	else {
		/* rc measures "distance" of color from red */
		double rc = (double)(maximum - red())   / (double)(maximum - minimum);
		double gc = (double)(maximum - green()) / (double)(maximum - minimum);
		double bc = (double)(maximum - blue())  / (double)(maximum - minimum);

		if (red() == maximum)
			*h = bc - gc;		/* resulting color between yellow and magenta */
		else
		if (green() == maximum)
			*h = 2.0 + rc - bc;	/* resulting color between cyan and yellow */
		else
			*h = 4.0 + gc - rc;	/* resulting color between magenta and cyan */

		*h *= 60.0;			/* convert to degrees */

		if (*h < 0.0) *h += 360.0;	/* make non negative */
	}  /* chromatic case */
} // getHSV

/** setHSV - convert an HSV color to RGB
 * @param	h in [0..360]
 * @param	s in [0..1]
 * @param	v in [0..1], except if s = 0.0
 * @return	rgb value of color in dword
 */
void Color32::setHSV(double h, double s, double v)
{
	double r, g, b;

	if (s==0.0) {			/* achromatic color there is no hue */
		if (h==-1.0)		/* undefined */
			r = g = b = v;	/* this is the achromatic case */
		 else
			r = g = b = 0;	/* error if s = 0 and h has a value */
	} else {			/* chromatic color: there is a hue */
		if (h==360.0) h = 0.0;
		h /= 60.0;		/* h is now in [0,6] */

		int i = (int)h;		/* largest integer <= h */
		double f = h - (double)i;/* fractional part of h */
		double p = v * (1.0 - s);
		double q = v * (1.0 - s*f);
		double t = v * (1.0 - s*(1.0 - f));

		switch (i) {
			case  0:
				r = v;
				g = t;
				b = p;
				break;

			case  1:
				r = q;
				g = v;
				b = p;
				break;

			case  2:
				r = p;
				g = v;
				b = t;
				break;

			case  3:
				r = p;
				g = q;
				b = v;
				break;

			case  4:
				r = t;
				g = p;
				b = v;
				break;

			case  5:
				r = v;
				g = p;
				b = q;
				break;

			default:
				r = g = b = 0;
		}  /* end of switch */
	}

	red()   = int(r*255.0);
	green() = int(g*255.0);
	blue()  = int(b*255.0);
} // setHSV

/* =============================== Color ================================== */
/* ----------------------------------------------------------------- *
 *  setHSV(h,s,v);                                                 *
 *  Input:   h in [0,360] or undefined (-1), s and v in [0,1]        *
 * ----------------------------------------------------------------- */
void Color::setHSV(float h, float s, float v)
{
	if (Eq0(s, _CEPSILON)) {		// achromatic color there is no hue
		if (Eq(h, -1.0, _CEPSILON))	// undefined
			// this is the achromatic case
			_red = _green = _blue = v;
//		else
//			cerr << WARNINGSTR
//				<< "Color::setHSV invalid values" << endl;
	} else {				// chromatic color: there is a hue
		if (Eq(h, 360.0, _CEPSILON)) h = 0.0;
		h /= 60.0f;			// h is now in [0,6]

		int i = (int)h;			// largest integer <= h
		float f = h - (float)i;		// fractional ray of h
		float p = v * (1-s);
		float q = v * (1 - s*f);
		float t = v * (1 - s*(1-f));

		switch (i) {
			case  0:
				_red   = v;
				_green = t;
				_blue  = p;
				break;

			case  1:
				_red   = q;
				_green = v;
				_blue  = p;
				break;

			case  2:
				_red   = p;
				_green = v;
				_blue  = t;
				break;

			case  3:
				_red   = p;
				_green = q;
				_blue  = v;
				break;

			case  4:
				_red   = t;
				_green = p;
				_blue  = v;
				break;

			case  5:
				_red   = v;
				_green = p;
				_blue  = q;
				break;

			default: ;
//				cerr << WARNINGSTR <<
//				"Color::setHSV invalid values" << endl;
		}  /* end of switch */
	}
} // setHSV

/* ========================= HSV ================================== *

			      ^ Value
		       _______|______
		       \      |      /   Side face of hexcone HSV
			 \    |    /
			   \  |  /
			     \|/
		    Green          Yellow               /
			-------------                 /
		      / \           / \             /
		    /     \       /     \         /  ^  Hue
		  /         \   /         \           \
	  Cyan  /_____________x_____________\   Red  --------->
		\       White  \            /                Saturation
		  \         /    \        /
		    \     /        \    /
		      \ /            \/          Top face of hexcone HSV
		  Blue  ~~~~~~~~~~~~~~ Magenta    color model

 * ================================================================ */

/* ================================================================= *
 *  getHSV(float& h, float& s, float& v);                                *
 *                                                                   *
 *  Input:   r,g,b, each in [0,1]                                    *
 *  Output:  h in [0,360],  s and v in [0,1], except if s = 0        *
 *           then h is undefined which is defined as a constant -1   *
 *           outside the interval [0,360]                            *
 * ================================================================= */
void Color::getHSV(float& h, float& s, float& v)
{
	float maximum = Max(_red,_green,_blue);
	float minimum = Min(_red,_green,_blue);

	v = maximum;	/* value */

	if (!Eq0(maximum, _CEPSILON))
		s = (maximum - minimum) / maximum;	// saturation
	else
		s = 0;

	if (Eq0(s, _CEPSILON))
		h = -1;		// hue, undefined
	else {
		// rc measures "distance" of color from red
		float rc = (maximum - _red) / (maximum - minimum);
		float gc = (maximum - _green) / (maximum - minimum);
		float bc = (maximum - _blue)  / (maximum - minimum);

		if (Eq(_red, maximum, _CEPSILON))
			// resulting color between yellow and magenta
			h = bc - gc;
		else
		if (Eq(_green, maximum, _CEPSILON))
			// resulting color between cyan and yellow
			h = 2 + rc - bc;
		else
			// resulting color between magenta and cyan
			h = 4 + gc - rc;

		h *= 60;		// convert to degrees

		if (h < 0) h += 360;	// make non negative
	}  /* chromatic case */
} // getHSV

/* ============================== Color3D ================================= */
void Color3D::operator () (dword c)
{
	double h, s, v;

	color = c & 0xFFFFFF;
	color.getHSV(&h, &s, &v);
	dark.setHSV(h, s, (1.0 + v)/2.0);
	bright.setHSV(h, s, 1.0);
} /* operator () */
