/*
 * Copyright European Organization for Nuclear Research (CERN)
 * All rights reserved
 *
 * Please look at the supplied documentation for the user's
 * license
 *
 * Usrbin manipulator allowing for calculating ratios, sums etc. of usrbins
 *
 * Authors: Christian.Theis@cern.ch
 *          Vasilis.Vlachoudis@cern.ch
 *
 * Date:    09-Mar-2023
 */

#include "usrbin_manip.h"

using namespace std;

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

int main( /*int argc, char** argv */ )

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
	vector<string> fileNames;
	string   inputStr, opStr;
	int      detectorIndex = 1, nrOfFiles = 0;
	double   curWeight = 1.0;

	cout << "USRBIN manipulator" << endl;
	cout << "Version 1.0 / 09/03/2023" << endl;
	cout << "Authors: C. Theis, V. Vlachoudis" << endl;
	cout << "--------------------------------" << endl << endl;

	while ( true ) {
		cout << "How many files (>= 2) do you want to manipulate: ";
		cin >> nrOfFiles;
		if ( nrOfFiles > 1 ) break;
	}

	for ( int i = 1; i <= nrOfFiles; ++i ) {
		cout << "Enter usrbin filename " << i << " and it's associated weight (e.g. \"test.bnn 1.0\"): ";
		cin >> inputStr >> curWeight;
		fileNames.push_back( inputStr );
		_meshWeights.push_back( curWeight );
	}

	Usrbin inMesh;
	auto detNames = inMesh.getDetectorNames( fileNames[0].c_str() );
	cout << endl << "The data file includes " << detNames.size() << " data meshes." << endl;
	for ( auto curName : detNames ) {
		cout << detectorIndex++ << ".) " << curName << endl;
	}

	cout << endl << "Enter the index of the detector to manipulate (>= 1): ";
	cin >> detectorIndex;
	detectorIndex = std::max( 1, detectorIndex );

	if ( !LoadMeshes( fileNames, detectorIndex ) ) {
		cerr << "ABORT: could not read all meshes" << endl;
		return -1;
	}

	while ( true ) {
		cout << endl << "Enter the mathematical operator to be carried out as A # B where # is one of the following operators \"+ - * / |-|\" : ";
		cin >> opStr;
		if ( _manipulators.find( opStr ) == _manipulators.end() ) {
			cout << "Invalid selection!" << endl;
			continue;
		}
		else
			break;
	}

	auto functionOp = _manipulators[opStr];
	std::function<double( double a, double w1 )> errOp;
	if ( opStr != "|-|" )
		errOp = _variance["%"];		// define how the error propagation shall be calculated
	else
		errOp = _variance["0"];

	cout << endl << "Enter output filename: ";
	cin >> inputStr;
	cout << endl;

	if ( ManipulateMeshes( inputStr, functionOp, errOp, opStr ) ) {
		cout << "Successfully terminated!" << endl;
	}
	else {
		cout << "Sorry, we encountered a problem during the processing!" << endl;
		return -1;
	}

	return 1;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

bool LoadMeshes( const std::vector<std::string>& fileNames, int detectorIndex )

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{

	_meshes.clear();
	for ( const auto& curFileName : fileNames ) {
		std::unique_ptr<Usrbin> pCurUsrbin( new Usrbin() );
		if ( !pCurUsrbin->load( curFileName.c_str(), detectorIndex ) ) {
			cerr << "ERROR: Could not read " << curFileName << endl;
			return false;
		}
		_meshes.emplace_back( std::move(pCurUsrbin) );
	}

	if ( _meshes.empty() )
		return false;

	// validate if all meshes contain the same data layout
	int i = 0;
	for ( const auto& curMesh : _meshes ) {

		if ( !curMesh->hasData() ) {
			cerr << "ERROR: no data found in " << fileNames[i] << ". Did you specify the manipulation for a detector index that does not exist??" << endl;
			return false;
		}

		// check type
		if ( curMesh->type() != _meshes[0]->type() ) {
			cerr << "ERROR: mismatching type [" << (int)_meshes[0]->type() << "] of mesh 1 " << fileNames[0] << " and " << fileNames[i] << "[" << (int)curMesh->type() << "]" << endl;
			return false;
		}

		// check spatial layout
		if ( _meshes[0]->nx != curMesh->nx || _meshes[0]->ny != curMesh->ny || _meshes[0]->nz != curMesh->nz ) {
			cerr << "ERROR: mismatching number of bins between " << fileNames[0] << " and " << fileNames[i] << endl;
			return false;
		}
		else if ( _meshes[0]->dx != curMesh->dx || _meshes[0]->dy != curMesh->dy || _meshes[0]->dz != curMesh->dz ) {
			cerr << "ERROR: mismatching binsize between " << fileNames[0] << " and " << fileNames[i] << endl;
			return false;
		}
		else if ( _meshes[0]->xlow !=  curMesh->xlow  || _meshes[0]->ylow  != curMesh->ylow  || _meshes[0]->zlow  != curMesh->zlow ||
							_meshes[0]->xhigh != curMesh->xhigh || _meshes[0]->yhigh != curMesh->yhigh || _meshes[0]->zhigh != curMesh->zhigh ) {
			cerr << "ERROR: mismatching low/high coordinates between " << fileNames[0] << " and " << fileNames[i] << endl;
			return false;
		}

		++i;
	}

	return true;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

bool ManipulateMeshes( const std::string& outputFileName, std::function<double( double a, double b, double w1, double w2 ) > mathOp,
											 std::function<double( double a, double w1) > varOp, const std::string& opStr )

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
	Usrbin   outMesh;
	double   resVal = 0, nextVal = 0, curWeight = 0, resRelErr = 0, curVal = 0, curErr = 0;
	bool     bVerbose = false;
	ofstream debugStr;

	if( bVerbose )
		debugStr.open( "logfile.txt" );

	outMesh.create( 1, _meshes[0]->type(), _meshes[0]->score(), _meshes[0]->xlow, _meshes[0]->xhigh, _meshes[0]->nx,
									_meshes[0]->ylow, _meshes[0]->yhigh, _meshes[0]->ny,
									_meshes[0]->zlow, _meshes[0]->zhigh, _meshes[0]->nz );


	double totalWeight = accumulate( _meshWeights.begin(), _meshWeights.end(), 0.0 );

	for ( int k = 0; k < _meshes[0]->nz; ++k ) {
		for ( int j = 0; j < _meshes[0]->ny; ++j ) {
			for ( int i = 0; i < _meshes[0]->nx; ++i ) {

				resVal    = _meshes[0]->get( i, j, k ) * _meshWeights[0] / totalWeight;
				resRelErr = varOp( _meshes[0]->get( i, j, k, true ), _meshWeights[0] / totalWeight );

				bVerbose && debugStr << _meshes[0]->get( i, j, k ) << " [+/- " << _meshes[0]->get( i, j, k, true ) * _meshWeights[0] * 100 << " %]  * " << _meshWeights[0] / totalWeight;

				for ( size_t c = 1; c < _meshes.size(); ++c ) {
					nextVal = _meshes[c]->get( i, j, k);
					curWeight = _meshWeights[c];

					// manipulate value in sequence
					bVerbose && debugStr << " " << opStr << " " << nextVal << "[+/- " << _meshes[c]->get( i, j, k, true ) * 100.0 << "%] * " << curWeight / totalWeight;
					resVal = mathOp( resVal, nextVal, 1.0, curWeight / totalWeight );		// do the math including weights

					// propagate uncertainty
					resRelErr += varOp( _meshes[c]->get( i, j, k, true ), curWeight / totalWeight );
				}
				outMesh.set( i, j, k, resVal );

				// final uncertainty
				resRelErr = sqrt( resRelErr );
				outMesh.set( i, j, k, resRelErr, true);

				bVerbose && debugStr << " = " <<  resVal << " +/- " << resRelErr * 100.0  << " %" << endl;
			}
		}
	}

	return outMesh.save( outputFileName.c_str() );
}
/* vim: ts=2 */
