/*
 * Copyright European Organization for Nuclear Research (CERN)
 * All rights reserved
 *
 * Please look at the supplied documentation for the user's
 * license
 *
 * Author:	Vasilis.Vlachoudis@cern.ch
 * Date:        19 Feb 2013
 */

#include <string>
#include <fstream>
#include <iostream>

#include <stdio.h>
#include <stdlib.h>

#include "os.h"
#include "nox.h"
#include "tool/token.h"

#define _DRAW_BBOX false

using std::cout;
using std::cerr;
using std::endl;

const char* SEP = " \t";

/** NoxGeometry */
NoxGeometry::NoxGeometry(const char* filename)
	: kernel(geometry),
	  viewer(geometry,kernel),
	  dump(false)
{
	geometry.backgroundColor(0x707070);
	if (filename) load(filename);
} // NoxGeometry

/** parse matrix */
void NoxGeometry::parseMatrix(Token& token, Matrix4& matrix)
{
	matrix.identity();
	for (int row=0; row<4; row++)
		for (int col=0; col<4; col++) {
			if (!token) break;
			matrix(row,col) = token.getNumber();
		}
} // parseMatrix

/** parseBody */
GBody* NoxGeometry::parseBody(Token& token)
{
	GBody* body;
	int n=0;
	std::string type = token.getIdent();
	std::string name = token.string();	// accept string or number
	token.next();

	double what[30];
	memset(what, 0, sizeof(what));

	if (type == "VOXELS") {
		if (!geometry.voxel.load(name.c_str())) {
			cerr << "ERROR Unable to load voxel " << name << endl;
		}
		geometry.voxel.xlow = token.getNumber();
		geometry.voxel.ylow = token.getNumber();
		geometry.voxel.zlow = token.getNumber();
		geometry.voxel.calcLimits();

		// create a voxel body
		body = geometry.addBody("VOXEL", "RPP");
		what[n++] = geometry.voxel.xlow;
		what[n++] = geometry.voxel.xhigh;
		what[n++] = geometry.voxel.ylow;
		what[n++] = geometry.voxel.yhigh;
		what[n++] = geometry.voxel.zlow;
		what[n++] = geometry.voxel.zhigh;

	} else {
		body = geometry.addBody(name, type);
		while (token and token.is(Token::Type::NUMBER))
			what[n++] = token.getNumber();
	}

	try {
		body->set(n,what);
	} catch (std::string err) {
		cerr << err << endl;
	}
	body->create();
	if (dump) cout << "Body: " << name << " " << type << endl;
	return body;
} // parseBody

/** parseRegion */
GRegion* NoxGeometry::parseRegion(Token& token)
{
	// should accept, identifier, string and integer
	GRegion* region = geometry.addRegion(token.string());
	token.next();
	if (dump) cout << "Region: " << region->name() << endl;

	region->type((GRegion::Type)token.getInteger());
	token.next();
	//int rotdefi = atoi(strtok(NULL,SEP));// FIXME rotdefi ignore for the moment!!!!!
	//region->matrix = atoi(strtok(NULL,SEP));	FIXME
	region_material_t regmat;
	regmat.region_id = region->id();
	regmat.color = token.getInteger();
	regmat.alpha = token.getInteger();
	region_materials.push_back(regmat);
#if _DRAW_BBOX
	region->show = BIT_BBOX;
#endif
	if (!strcmp(token.ident(), "NORM")) {
		if (geometry.parseRegion(*region, token))
			exit(2);
	} else
		assert(0);
	return region;
} // parseRegion

/** parseVoxel */
void NoxGeometry::parseVoxel(Token& token)
{
	if (token.getIdent() == "color") {
		int i = token.getInteger();
		dword color = (dword)token.getInteger();
		viewer.voxel().color(i-1, color);
	}
} // parseVoxel

/** parse */
bool NoxGeometry::parse(std::ifstream& input)
{
	Matrix4 matrix;			// current matrix
	bool hasMatrix = false;
	GRegion* region = NULL;		// current region

	std::string line;
	while (std::getline(input, line)) {
		if (line.empty() or line[0]=='#') continue;

		std::stringstream stream(line);
		Token token(stream);

		// continuation lines
		if (line[0]==' ' or line[0]=='\t') {
			// continuation line
			if (region!=NULL) {
				// add to the previous region
				if (geometry.parseRegion(*region, token))
					exit(2);
			} else {
				cerr << "Invalid continuation line: " << line << endl;
				//cerr << "token=" << token << endl;
				//cerr << "region=" << region << endl;
			}
			continue;
		}

		// initialize current values
		region = NULL;

		token.next();
		if (!strcmp(token.ident(), "rotdefi")) {
			token.next();
			//int idx = atoi(strtok(NULL,SEP));
			//parseMatrix(matrix);
			//geometry.rotdefi(idx,matrix);
		} else
		if (!strcmp(token.ident(), "matrix")) {
			token.next();
			parseMatrix(token, matrix);
			hasMatrix = true;
			if (dump) cout << "\tMatrix" << endl;
		} else
		if (!strcmp(token.ident(), "body")) {
			// set additional identifier symbols
			token.setAlpha(".@/",Token::AlphaNumeric::SUFFIX);
			token.next();
			auto body = parseBody(token);
			if (hasMatrix) {
				body->matrix(matrix);
				body->transform();
				hasMatrix = false;
			}
		} else
		if (!strcmp(token.ident(), "region")) {
			token.next();
			region = parseRegion(token);
			if (hasMatrix) {
				region->matrix(matrix);
				hasMatrix = false;
			}
		} else
		if (!strcmp(token.ident(), "voxel")) {
			token.next();
			parseVoxel(token);
		} else
		if (!strcmp(token.ident(), "viewmatrix")) {
			token.next();
			parseMatrix(token, matrix);
			if (dump) cout << "View Matrix:" << endl << matrix << endl;
			viewer.matrix(matrix);
		} else
		if (!strcmp(token.ident(), "extends")) {
			token.next();
			double x1 = token.getNumber();
			double y1 = token.getNumber();
			double x2 = token.getNumber();
			double y2 = token.getNumber();
			if (dump) cout	<< "Extends [" << x1 << ", " << y1
					<< "] - [" << x2 << ", " << y2 << "]" << endl;
			viewer.window(x1,y1,x2,y2);
		} else
		if (!strcmp(token.ident(), "origin")) {
			token.next();
			double x = token.getNumber();
			double y = token.getNumber();
			double z = token.getNumber();
			if (dump) cout	<< "Origin [" << x << ", " << y
					<< ", " << z << "]" << endl;
			viewer.origin(x,y,z);
		} else
		if (!strcmp(token.ident(), "zoom")) {
			token.next();
			double z = token.getNumber();
			if (dump) cout	<< "Zoom " << z << endl;
			viewer.view().zoom(z);
		} else
		if (!strcmp(token.ident(), "mode")) {
			token.next();
			if (!strcmp(token.ident(), "fluka"))
				geometry.mode(Geometry::Mode::Fluka);
			else
			if (!strcmp(token.ident(), "moira"))
				geometry.mode(Geometry::Mode::Moira);
			else
			if (!strcmp(token.ident(), "mcnp"))
				geometry.mode(Geometry::Mode::Mcnp);
			else
			if (!strcmp(token.ident(), "phits"))
				geometry.mode(Geometry::Mode::Phits);
			else
			if (!strcmp(token.ident(), "penelope"))
				geometry.mode(Geometry::Mode::Penelope);
			else {
				cerr << "invalid mode '" << token.ident() << "'" << endl;
				return false;
			}
		} else {
			cerr << "Skipping unknown token " << token.string() << endl;
			return false;
		}
	}
	return true;
} // parse

/** assignRegionMaterials */
void NoxGeometry::assignRegionMaterials()
{
	// assign region materials
	for (auto regmat : region_materials) {
		VRegion* region = kernel.getRegion(regmat.region_id);
		if (region) {
			region->color(regmat.color);
			region->alpha(regmat.alpha);
		}
	}
} // assignRegionMaterials

/** load */
bool NoxGeometry::load(const char* filename)
{
	try {
		std::ifstream input(filename);
		if (input.fail()) return false;
		parse(input);
		input.close();
	} catch (std::string ex) {
		cerr << "Error reading file '" << filename << '\'' << endl;
		cerr << "Exception: " << ex << endl;
		return false;
	}
	geometry.derive();
	kernel.derive();
	assignRegionMaterials();	// must be called after derive
	return true;
} // load

/** cleanup */
void NoxGeometry::cleanup()
{
	geometry.cleanup();
} // cleanup
